/* -------------------- code.c ------------------------- */
#include <stdio.h>
#ifdef NEED_MALLOC
#include <malloc.h>
#endif
#include "common.h"
#include "code.h"

	/* global variables for code */
mnemonic mntbl[] = {            /* table for mnemonic code */
    { "LIT", O_LIT }, { "OPR", O_OPR }, { "LOD", O_LOD }, { "STO", O_STO },
    { "CAL", O_CAL }, { "INT", O_INT }, { "JMP", O_JMP }, { "JPC", O_JPC },
    { "CSP", O_CSP }, { "LAB", O_LAB }, { "   ", O_BAD }
};

int label[CMAX];		/* label table for backpatching */
int cl = 0;			/* current label number using O_LAB */ 

instruction code[CMAX];		/* code table */
int cx = 1;			/* code index */

	/* functions for code generation */
int genlabel()      /* generate a label */
{
    if( ++cl > CMAX ) {
	error( "too many labels" );
	exit( 1 );
    }
    return cl;
}

codeptr nullcode()  /* make empty code */
{
    codeptr rp;	    /* return value */

    rp.head = rp.tail = NULL;
    return rp;
}

codeptr gen( f, l, a ) /* generate a single instruction */
  opecode f;
  int l, a;
{
    codestr *sp;
    codeptr rp;	       /* return value */

    if(( sp = (codestr *)malloc( sizeof(codestr) )) == NULL ) {
	perror( "malloc" );
	exit( 1 );
    }
    sp->inst.f = f;  sp->inst.l = l;  sp->inst.a = a;  sp->next = NULL;
    rp.head = rp.tail = sp;
    return rp;
}

codeptr concat2( bp, cp )   /* concatenate codes bp and cp. */
  codeptr bp;               /* modified concat(n, cp) for gcc. */
  codeptr cp;
{
    codeptr *ap;
    codeptr *rp;

    rp = &bp;
    ap = &cp;
    if( ap->head != NULL ) {      /* skip NULL code */
	if( rp->head == NULL ) {  /* skip NULL code */
	    rp = ap;
	} else {
	    rp->tail->next = ap->head;  rp->tail = ap->tail;
	}
    }
    return *rp;
}

codeptr printcode( cp ) /* print code represented as a list structure */
  codeptr cp;
{
    codestr *wp;
    int f, l, a;

    printf( "\n" );
    wp = cp.head;
    while( wp != NULL ) {
	f = wp->inst.f;  l = wp->inst.l;  a = wp->inst.a;
	printf( "( %s, %3d, %3d )\n", mntbl[f].sym, l, a );
	wp = wp->next;
    }
    printf( "\f" );
    return cp;
}

static void wrcode( c ) /* print code stored in array code[] */
  int c;
{
    int i, f, l, a;
    
    printf( "code table\n\n" );
    for( i=1; i<=c; i++ ) {
	f = code[i].f;  l = code[i].l;  a = code[i].a;
	printf( "%4d [ %s, %3d, %3d ]\n", i, mntbl[f].sym, l, a );
    }
    printf( "\f" );
}

codeptr linearize( cp ) /* transform list structure code cp to array code[] */
  codeptr cp;           /* also do backpatching */
{
    codestr *wp;
    int f, l, a;
    int c1, c2;

    cleararray( code, sizeof(code) );
    cleararray( label, sizeof(label)); /* label is the label table */
    if( cp.head == NULL ) return cp;
    for( wp = cp.head; wp != NULL; wp = wp->next ) {
	f = wp->inst.f;  l = wp->inst.l;  a = wp->inst.a;
	switch( f ) {
	case O_LAB:
	    if( label[a] == 0 ) { /* label 'a' was unused */
		label[a] = cx;    /* let it be defined. 
                                     let label[a] point to its address */
	    } else if( label[a] < 0 ) { /* label 'a' was used before def. */
		c1 = -label[a];         /* backpatching */
		while( c1 != 0 ) {      /* trace forward reference chain */
		    c2 = code[c1].a;
		    code[c1].a = cx;
		    c1 = c2;
		}
		label[a] = cx;
	    } else {
		error( "label already defined" );
		exit( 1 );
	    }
	    break; /* do not increment cx (delete LAB instruction) */
	case O_JMP:
	case O_JPC:
	case O_CAL:
	    code[cx].f = f;
	    code[cx].l = l;
	    if( label[a] <= 0 ) { /* label 'a' was undefined. 
                                     make forward reference chain */
		code[cx].a = -label[a];
		label[a] = -cx;
	    } else {              /* label 'a' was defined */
		code[cx].a = label[a];
	    }
	    cx++;
	    break;
	default:
	    code[cx].f = f;  code[cx].l = l;  code[cx].a = a;
	    cx++;
	    break;
	}
	if( cx >= CMAX ) {
	    error( "program is too large" );
	    exit( 1 );
	}
    }
    wrcode( cx - 1 );
    return cp;
}
