/* ----------------------- env.c ------------------------ */
#include <stdio.h>
#include <string.h>
#ifdef NEED_MALLOC
#include <malloc.h>
#endif
#include "env.h"

	/* global variable for environment (symbol table) */
tblstr table[TXMAX]; /* the symbol table in an array */

	/* functions for environment management */
envptr allocenv() /* make a structure of type envstr */
{
    envptr ep;

    if(( ep = (envptr)malloc( sizeof(envstr) )) == NULL ) {
	perror( "malloc" );
	exit( 1 );
    }
    return ep;
}

envptr nullenv() /* initialize environment (symbol table) */
{
    envptr initenv = allocenv();

    initenv->lev = -1;
    initenv->dx  = 3;
    initenv->cur = &table[1];
    initenv->top = &table[0];
    return initenv;
}

envptr newenv( outerenv ) /* allocate a structure envstr for the new scope */
  envptr outerenv;        /* and return it */
{
    envptr innerenv = allocenv();

    innerenv->lev = outerenv->lev + 1;
    if( innerenv->lev > LEVMAX ) {
	error( "too deep nested scope" );
	exit( 1 );
    }
    innerenv->dx  = 3;
    innerenv->cur = outerenv->top + 1;
    innerenv->top = outerenv->top;
    return innerenv;
}

envptr enter( kind, id, num, env ) /* enter a new entry */
  idtype kind;                     /* and return the env (symbol table) */
  int num;
  char *id;
  envptr env;
{
    tblptr tp;
    envptr ep = allocenv();

    ep->lev = env->lev;
    ep->dx  = env->dx;
    ep->cur = env->cur;
    ep->top = env->top +1;
    tp = ep->top;
    if( tp > &table[TXMAX] ) {
	error( "too many identifiers" );
	exit();
    }
    tp->name = id;
    tp->kind = kind;
    switch( kind ) {
    case CONSTANT:
	if( num > AMAX ) {
	    error( "this number is too large" );
	    num = 0;
	}
	tp->val = num;
	break;
    case VARIABLE:
	tp->level = ep->lev;
	tp->adr = ep->dx;
	ep->dx++;
	break;
    case PROC:
	tp->level = ep->lev;
	tp->adr = num;
	break;
    }
    return ep;
}

tblptr lookup( id, env ) /* search id in env (including global environment) */
  char *id;              /* return the entry found */
  envptr env;
{
    tblptr tp;

    table[0].name = id;		/* &table[0] == SENTINEL */
    for( tp = env->top; strcmp( tp->name, id ) != 0; tp-- );
    return tp;
}

bool exist( id, env )   /* check whether id is in current env */
  char *id;             /* return TRUE or FALSE */
  envptr env;
{
    int r;
    tblptr tp;

    for( tp=env->top; tp>=env->cur; tp-- ) {
	if( strcmp( tp->name, id ) == 0 )
	    return TRUE;	/* id in current env */
    }
    return FALSE;		/* id not in current env */
}

envptr printenv( env ) /* print environment (symbol table) */
  envptr env;
{
    tblptr tp;

    printf( "\n" );
    tp = &table[1];
    while( tp <= env->top ) {
	printf( "%-15s", tp->name );
	switch( tp->kind ) {
	case CONSTANT:
	    printf( "const  %8d", tp->val );
	    break;
	case VARIABLE:
	    printf( "var    %8d %8d", tp->level, tp->adr );
	    break;
	case PROC:
	    printf( "proc   %8d %8d", tp->level, tp->adr );
	    break;
	}
	printf( "\n" );
	tp++;
    }
    printf( "\n" );
    return env;
}
