/* -------------------------- inter.c ------------------------- */
#include <stdio.h>
#include "code.h"
    
	/* global variables for interpreter */
#define STACKSIZE 500
int s[STACKSIZE]; /* run-time stack */
int p, b, t;      /* program counter, pointer to stack, top of stack */

void interpreter()        /* interpreter */
{
    instruction i;
    opecode f; /* operation code */
    int l, a;  /* l = level difference or 0
                  a = displacement or sub-operation or number */
    printf( "start PL/0\n" );
    t = 0;  b = 1;  p = 1;
    s[1] = s[2] = s[3] = 0; /* static link, dynamic link, ret. addr. of main */
    do {
	i = code[p++]; /* get an instruction */
	f = i.f;  l = i.l;  a = i.a;
	switch( f ) { /* switch by op-code */
	case O_LIT : s[++t] = a;  break;
	case O_OPR : 
	    switch( (oprcode)a ) { /* P_RET = return from procedure */
	    case P_RET: t = b - 1;  p = s[t+3];	 b = s[t+2];  break;
	    case P_NEG:	   ;  s[t] = -s[t]		   ;  break;
	    case P_ADD: --t;  s[t] = s[t] + s[t+1]	   ;  break; 
	    case P_SUB: --t;  s[t] = s[t] - s[t+1]	   ;  break; 
	    case P_MUL: --t;  s[t] = s[t] * s[t+1]	   ;  break; 
	    case P_DIV: --t;  s[t] = s[t] / s[t+1]	   ;  break; 
	    case P_ODD:	   ;  s[t] = s[t] % 2		   ;  break;
	    case P_EQ:	--t;  s[t] = ( s[t] == s[t+1] )	   ;  break;
	    case P_NE:	--t;  s[t] = ( s[t] != s[t+1] )	   ;  break;
	    case P_LT:	--t;  s[t] = ( s[t] <  s[t+1] )	   ;  break;
	    case P_GE:	--t;  s[t] = ( s[t] >= s[t+1] )	   ;  break;
	    case P_GT:	--t;  s[t] = ( s[t] >  s[t+1] )	   ;  break;
	    case P_LE:	--t;  s[t] = ( s[t] <= s[t+1] )	   ;  break;
	    }
	    break;
	case O_LOD: s[++t] = s[base(l)+a];  break;
	case O_STO: s[base(l)+a] = s[t--];  break;
	case O_CAL: s[t+1] = base(l) /* static link */ ;
                    s[t+2] = b /* dynamic link */; s[t+3] = p /* ret. addr. */;
		    b = t + 1;	p = a;	break;
	case O_INT: t += a;  break;
	case O_JMP: p = a;  break;
	case O_JPC: if( s[t--] == 0 ) p = a;  break;
	case O_CSP:
	    switch( (cspcode)a ) {
	    case RDI: if( scanf("%d",&s[++t])!=1 ) printf( "error read\n" );
                      break; /* read(var) */
	    case WRI: printf( "%10d", s[t--] );  break; /* write(exp) */
	    case WRL: putchar( '\n' );	break; /* writeln */
	    }
	    break;
	}
    } while( p != 0 );
    printf( "end PL/0\n" );
}

int base( l ) /* trace static link l times, where l is the difference */
  int l;      /* of current and referenced static nesting levels */
{
    int b1;

    b1	= b; /* b points to the area of static link */
    while( l>0 ) {
	b1 = s[b1];
	l--;
    }
    return b1;
}
