/* functions for handle attriubtes
   Copyright (C) 1991-1993 Masataka Sassa, Kazuhiro Kuroishi, Teruhisa Hirai.
   Copyright (C)      1994,1995                               Yoshiki Ohshima.
     
This file is part of Rie, A Compiler Compiler Based on ECLR-attributed Grammar.

Rie is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

Rie is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Bison; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

Bug reports, fixes or ideas are welcomed.
		Address: Yoshiki Ohshima
			 Tokyo Institute of Technology,
			 Department of Information Science,
			 Ookayama, Meguro-ku, Tokyo, 152 Japan.

		E-mail: rie-comments@is.titech.ac.jp			*/


#include <stdio.h>
#include "attribute.h"
#include "error.h"
#include "gram.h"
#include "new.h"
#include "symtab.h"
#include "system.h"

extern semantic_rule *first_semantic_rule;	/* reader.c  */
extern semantic_condition *first_condition;	/* reader.c  */

#if (defined(__STDC__) || defined(__GNUC__))

extern void output_semantic_condition(FILE *, semantic_condition *); /* outputsem.c */

#ifdef DEBUG
extern void print_pure(pure *);		/* pure.c  */
#endif

#else	/* not ANSI C nor GNU C  */

extern void output_semantic_condition(); /* outputsem.c */

#ifdef DEBUG
extern void print_pure();		/* pure.c  */
#endif

#endif	/* not ANSI C nor GNU C  */


int ninherits;
int nsynthesizes;
int nlocals;
int nEC;
int nlocal_classes;
semantic_rule **inh_table;	/* [0..nitems-1], inh. semrules at the item  */
semantic_rule **synt_table;	/* [1..nrules]  , synt.semrules at the rule  */
semantic_rule **local_table;	/* [0..nitems-1], loc. semrules at the item  */
semantic_condition **condition_table;
				/* [1..nrules], sem. conditions at the rule  */
short *irule;			/* [0..nitems-1], "item no->rule no" array.  */
short *rrhsend;			/* [1..nrules],   "rule no->item no" array.  */

#if (defined(__STDC__) || defined(__GNUC__))

void set_semantic_tables(void);
void free_semantic_tables(void);
void allocate_EC(void);
void allocate_local_classes(void);
void set_irule(void);
void free_irule(void);
void set_rrhsend(void);
void free_rrhsend(void);

#ifdef DEBUG
void print_semantic_tables(void);
void print_expressions(semantic_expression *);
void print_semrules_d(semantic_rule *);
void print_semrule_d(semantic_rule *);
void print_all_semantic_rules(void);
void print_semcond_d(semantic_condition *);
#endif

#else	/* not ANSI C nor GNU C  */

void set_semantic_tables();
void free_semantic_tables();
void allocate_EC();
void allocate_local_classes();
void set_irule();
void free_irule();
void set_rrhsend();
void free_rrhsend();

#ifdef DEBUG
void print_semantic_tables();
void print_expressions();
void print_semrules_d();
void print_semrule_d();
void print_all_semantic_rules();
void print_semcond_d();
#endif

#endif	/* not ANSI C nor GNU C  */



/*
    Distribute all semantic rules chained from 'first_semantic_rule' into
  'inh_table[]', 'synt_table[]', 'local_table[]' according to its attribute
  class.  And all semantic conditions chained from 'first_condition' into
  'condition_table[]'.
*/

void
set_semantic_tables()
{
  register semantic_rule *rule, *next, *rp;
  register semantic_condition *cond;
  register attribute *ap;
  register int ruleno;

  /* semantic rules  */
  /* number of attributes.  */
  nlocals = 0;

  inh_table = NEW2(nitems, semantic_rule *);
  synt_table = NEW2(nrules, semantic_rule *) - 1;
  local_table = NEW2(nitems, semantic_rule *);
  
  rule = first_semantic_rule;

  while (rule)
    {
      /* skip and free terminators which indicate the end of each rule  */
      while (rule && rule->expression == NULL)
	{
	  next = rule->link;
	  FREE(rule);
	  rule = next;
	}

      while (rule && rule->expression)
	{
	  next = rule->link;

	  ap = rule->attribute;
	  switch (ap->pure->class)
	    {
	    case INHERIT:
	      /* 'ninherits' is counted at parse_token_decl().  */
	      rule->link = inh_table[ap->offset];
	      inh_table[ap->offset] = rule;
	      break;
	      
	    case SYNTHESIZE:
	      /* 'nsyntesizes' is counted at parse_token_decl().  */
	      ruleno = - ritem[ap->offset];
	      if (ruleno < 0)
		internal("at set_semantic_tables()");
	      
	      rule->link = synt_table[ruleno];
	      synt_table[ruleno] = rule;
	      break;

	    case LOCAL:
	      nlocals++;
	      /* apppend to the tail */
	      rule->link = NULL;
	      rp = local_table[ap->offset];
	      if (rp == NULL)
		{
		  local_table[ap->offset] = rule;
		}
	      else
		{
		  while(rp->link)
		    rp = rp->link;
		  rp->link = rule;
		}
	      break;

	    default:
	      internal("at set_semantic_tables()");
	    }

	  rule = next;
	}
    }

  /* 'first_semantic_rule' is no more useful.  */
  first_semantic_rule = NULL;


  /* semantic conditions  */

  condition_table = NEW2(nrules, semantic_condition *) - 1; /* [1..nruls-1] */

  cond = first_condition;
  while (cond)
    {
      semantic_condition *last = NULL;
      
      /* skip and free terminators which indicate the end of each condition. */
      while (cond && cond->kind == CTERMINATOR)
	{
	  semantic_condition *next;

	  next = cond->link;
	  FREE(rule);
	  cond = next;
	}

      if (cond)
	{
	  /* "cond->kind != CTERMINATOR"  */
	  ruleno = - ritem[cond->base_offset];
	  if (ruleno < 0)
	    internal("at set_semantic_tables()");
	  condition_table[ruleno] = cond;
	}
      
      /* set the 'link' member of "NULL" before a terminator  */
      for (; cond && cond->kind != CTERMINATOR; cond=cond->link)
	last = cond;
      if(last != NULL)
        last->link = NULL;
      /* "cond == NULL || cond->kind == CTERMINATOR"  */
  }
}



/* free semantic tables.  */

void
free_semantic_tables()
{
    FREE(inh_table);
    FREE(&synt_table[1]);
    FREE(local_table);
}



/* allocate EC to inherited attributes which is not appered any %equiv.  */

void
allocate_EC()
{
  register bucket *bp;
  register pure_list *list;
  register pure *purep;

#ifdef DEBUG
  printf("\nAllocating EC...\n");
#endif /* DEBUG */

  bp = firstsymbol;
  while(bp)
    {
      list = bp->pure_list;
      while (list)
        {
          purep = list->pure;
          if (purep->class == INHERIT && purep->EC == EC_UNDEF)
	    {
	      purep->EC = nEC++;
#ifdef DEBUG
	      print_pure(purep);
#endif /* DEBUG */
	    }
          list = list->link ;
        }
      bp = bp->next;
    }
#ifdef DEBUG
  printf("finish allocating EC. number of EC is %d\n\n", nEC);
#endif /* DEBUG */
}


/* allocate local classes for local attributes.  */

void
allocate_local_classes()
{
  register semantic_rule **tp, **tpend, *rp;

  /* number of local classes.  */
  nlocal_classes = 0;

  tp = local_table;
  tpend = local_table + nitems;

  while (tp < tpend)
    {
      rp = *tp++;
      while (rp)
	{
	  if (rp->attribute->pure->EC == EC_UNDEF)
	    rp->attribute->pure->EC = nlocal_classes++;
	  rp = rp->link;
	}
    }
#ifdef DEBUG
  printf("\nnlocal_classes = %d\n\n", nlocal_classes);
#endif
}


/*
  Before call, 'rrhsend[]' should be set, i.e. 'set_rrhsend()' called.
  'iruls[]' is an conversion array from item no to rule no.
*/

void
set_irule()
{
  register short ruleno;
  register short *ip;
  register int nitems_of_the_rule;
  
  irule = NEW2(nitems, short);	/* allocate memory */

  ip    = irule;
  ruleno= 1;
  nitems_of_the_rule = rrhsend[ruleno] - rrhs[ruleno] + 1;
  for (; ruleno <= nrules; ++ruleno,nitems_of_the_rule=rrhsend[ruleno]-rrhs[ruleno]+1)
    for (; 0 < nitems_of_the_rule; --nitems_of_the_rule)
      *ip++ = ruleno;
}


void
free_irule()
{
  FREE(irule);
}



/* set 'rrhsend[]'.  */

void
set_rrhsend()
{
  register int rule;

  rrhsend = NEW2(nrules, short) - 1;

  for (rule=1; rule < nrules; rule++)
    rrhsend[rule] = rrhs[rule+1] - 1;

  rrhsend[nrules] = nitems - 1;
}


/* free 'rrhsend[]'.  */

void
free_rrhsend()
{
  FREE(&rrhsend[1]);
}


#ifdef DEBUG

/* print all semantic rules in 'semrule[]'  */

void
print_semantic_tables()
{
  register semantic_rule *rp, **tp, **tpend;
  semantic_condition **cp, **cpend;

  /* print 'inh_table[]'  */
  printf("\nAll semantic rules in 'inh_table[]'\n");
  tp = inh_table;
  tpend = inh_table + nitems;
  while (tp < tpend)
    {
      printf("inh_table[%d] :\n", tp - inh_table);
      rp = *tp++;
      while (rp)
	{
	  printf("\t");
	  print_semrule_d(rp);
	  rp = rp->link;
	}
      printf("\n");
    }


  /* print 'synt_table[]'  */
  printf("\nAll semantic rules in 'synt_table[]'\n");
  tp = synt_table + 1;		/* index begins from 1  */
  tpend = synt_table + 1 + nrules;
  while (tp < tpend)
    {
      printf("synt_table[%d] :\n", tp - synt_table);
      rp = *tp++;
      while (rp)
	{
	  printf("\t");
	  print_semrule_d(rp);
	  rp = rp->link;
	}
      printf("\n");
    }


  /* print 'local_table[]'  */
  printf("\nAll semantic rules in 'local_table[]'\n");
  tp = local_table;
  tpend = local_table + nitems;
  while (tp < tpend)
    {
      printf("local_table[%d] :\n", tp - local_table);
      rp = *tp++;
      while (rp)
	{
	  printf("\t");
	  print_semrule_d(rp);
	  rp = rp->link;
	}
      printf("\n");
    }


  /* print 'condition_table[]'  */
  printf("\nAll semantic conditions in 'condition_table[]'\n");
  cp = condition_table + 1;
  cpend = condition_table + 1 + nrules;
  for (; cp < cpend; ++cp)
    {
      semantic_condition *condp;
      
      printf("condition_table[%d] :\n", cp - condition_table);
      condp = *cp;
      for (; condp; condp=condp->link)
	{
	  printf("\t");
	  output_semantic_condition(stdout, condp);
	}
      printf("\n");
    }
}



/* print expressions.  */

void
print_expressions(exp)
     semantic_expression *exp;
{
  register attribute *attrp;
  register pure *purep;
  
  while (exp)
    {
      attrp = exp->pointer.attribute;
      purep = attrp->pure;
      switch (exp->kind)
	{
	case KINHERIT:
	  printf(" %s.%s[%d]{%d}"
		 , attrp->tag, purep->name, attrp->offset, purep->EC);
	  break;

	case KSYNTHESIZE:
	  printf(" %s.%s[%d]", attrp->tag, purep->name, attrp->offset);
	  break;

	case KLOCAL:
	  printf(" %s[%d]{%d}", purep->name, attrp->offset, purep->EC);
	  break;

	case KCODE:
	  printf(" %s", exp->pointer.code);
	  break;

	default:
	  printf(" {unknown expression}");
	  break;
	}

      exp = exp->next;
    }
  printf("\n");
}



/* print semantic rules.  */


void
print_semrules_d(rule)
     semantic_rule *rule;
{
  while (rule)
    {
      if (rule->attribute)
	  print_semrule_d(rule);
      rule = rule->link;
    }
}



/* print a semantic rule.  */


void
print_semrule_d(rule)
     semantic_rule *rule;
{
  register attribute *attrp;
  register pure *purep;

  attrp = rule->attribute;
  purep = attrp->pure;
  switch (purep->class)
    {
    case INHERIT:
      printf("%s.%s[%d]{%d} ="
	     , attrp->tag, purep->name, attrp->offset, purep->EC);
      break;

    case SYNTHESIZE:
      printf("%s.%s[%d] =", attrp->tag, purep->name, attrp->offset);
      break;

    case LOCAL:
      printf("%s[%d]{%d} =", purep->name, attrp->offset, purep->EC);
      break;

    default:
      printf("{unknown} =");
      break;
    }

  print_expressions(rule->expression);
}



/* print all semantic rules listed from 'first_semantic_rule'  */


void
print_all_semantic_rules()
{
  register semantic_rule *rule;
  register int ruleno;	/* rule number is type int ?  */
  
  ruleno = 1;

  printf("\nAll semantic rules...\n");
  printf("semantic rules of production rule %d\n", ruleno);
  
  rule = first_semantic_rule;
  while (rule)
    {
      if (rule->expression == NULL)
	  /* end of one rule  */
	  printf("\nsemantic rules of production rule %d\n", ++ruleno);
      else
	{
	  printf("\t");
	  print_semrule_d(rule);
	}
      
      rule = rule->link;
    }
  
}


/* print 'first_condition'  */

void
print_first_condition()
{
  semantic_condition *cp;

  printf("'first_condition'...\n");
  
  cp = first_condition;
  for (; cp; cp=cp->link)
    {
      if (cp->kind != CTERMINATOR)
	output_semantic_condition(stdout, cp);
      else
	printf("---\n");
    }

  printf("over.\n\n");
}


/* functions for debug.  */
#endif
