/* define attribute types
   Copyright (C) 1991-1993 Masataka Sassa, Kazuhiro Kuroishi, Teruhisa Hirai.
   Copyright (C)      1994,1995                               Yoshiki Ohshima.
     
This file is part of Rie, A Compiler Compiler Based on ECLR-attributed Grammar.

Rie is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

Rie is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Bison; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

Bug reports, fixes or ideas are welcomed.
		Address: Yoshiki Ohshima
			 Tokyo Institute of Technology,
			 Department of Information Science,
			 Ookayama, Meguro-ku, Tokyo, 152 Japan.

		E-mail: rie-comments@is.titech.ac.jp			*/

#ifndef _ATTRIBUTE_H
#define _ATTRIBUTE_H

#include "pure.h"

/*
    Here define some C types for semantic rules, attributes, and semantic
  conditions.
    Type 'attribute' express an attribute belonged in a symbol. Which holds
  a symbol name as 'tag', a location of appeared as 'offset', an index of
  symbol among the corresponding production rule as 'index', and a *pure*
  as 'pure'(see pure.h).
 
    Type 'production_rules' express a production rule. Which is resemble as
  'bucket', but differs in holding a list of 'attribute' not holding
  'pure's as 'bucket'. This type and variables are used only by Rie not
  by Bison original.
 
*/

/*  kind of semantic expression  */

#define KUNKNOWN	0
#define KINHERIT	1
#define KSYNTHESIZE	2
#define KLOCAL		3
#define KCODE		4
#define KSYNTHESIZE_LHS 10

/* kind of semantic condition  */

#define CTERMINATOR	0
#define CCONDITION	1
#define CMESSAGE	2
#define CACTION		3
#define CDEBUG		4

/*  undefined offset for local attribute  */

#define UNDEF		-1

/*  truth values  */
#ifndef TRUE
#define TRUE		1
#endif /* TRUE */
#ifndef FALSE
#define FALSE		0
#endif /* FALSE */ 


/* used for 'flag' member of 'semantic_rule_list' type  */
#define ECLR_OK		0
#define ECLR_VIOLATION	1

typedef
  struct attribute
    {
      char *tag;
      short offset;
      short index;
      pure *pure;
    }
  attribute;


typedef
  struct attribute_list
    {
      struct attribute_list *link;

      attribute *attribute;
    }
  attribute_list;


typedef
  union pointers
    {
      attribute *attribute;
      char *code;
    }
  pointers;


typedef
  struct semantic_expression
    {
      struct semantic_expression *next;

      int kind;
      union pointers pointer;
    }
  semantic_expression;


typedef
  struct semantic_rule
    {
      struct semantic_rule *link;

      int lineno;
      semantic_expression *expression;
      attribute  *attribute;
    }
  semantic_rule;


typedef
  struct semantic_rule_list
    {
      struct semantic_rule_list *link;

      int flag;
      semantic_rule *rule;
    }
  semantic_rule_list;


typedef
  struct production_rule
    {
      struct production_rule *next;

      char *tag;
      attribute_list *attribute_list;
    }
  production_rule;


typedef
  struct semantic_expression
  conditional_expression;	/* type aliase */


typedef
  struct semantic_condition
    {
      struct semantic_condition *link;
      int lineno;
      int kind;
      int base_offset;
      conditional_expression *expression;
    }
  semantic_condition;

#endif /* _ATTRIBUTE_H  */
