/* Error message manager for Rie,
   Copyright (C) 1991-1993 Masataka Sassa, Kazuhiro Kuroishi, Teruhisa Hirai.
   Copyright (C)      1994,1995                              Yoshiki Ohshima.
     
This file is part of Rie, A Compiler Compiler Based on ECLR-attributed Grammar.

Rie is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

Rie is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Bison; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

Bug reports, fixes or ideas are welcomed.
		Address: Yoshiki Ohshima
			 Tokyo Institute of Technology,
			 Department of Information Science,
			 Ookayama, Meguro-ku, Tokyo, 152 Japan.

		E-mail: rie-comments@is.titech.ac.jp			*/

#include <stdio.h>
#include "attribute.h"
#include "files.h"
#include "system.h"

extern int lineno;		/* reader.c  */

extern void done();		/* files.c  */

int nerrors;

/* print warning message. DON'T EXIT */

void warning();
void warning_with_line();
void recovery();
void recovery_with_line();
void recovery_noline();
void clear_recovery();
void barrier_recovery();
void internal();

void
warning(fmt, x1, x2, x3, x4, x5, x6, x7, x8)
     char *fmt;
{
  warning_with_line(lineno, fmt, x1, x2, x3, x4, x5, x6, x7, x8);
}


void
warning_with_line(lno, fmt, x1, x2, x3, x4, x5, x6, x7, x8)
     int lno;
     char *fmt;
{
  if (infile != 0)
    {
      fprintf(stderr, "%s:%d: warning: ", infile, lno);
      fprintf(stderr, fmt, x1, x2, x3, x4, x5, x6, x7, x8);
      putc('\n', stderr);
    }
}


/* print error message for recovery.  */

void
recovery(fmt, x1, x2, x3, x4, x5, x6, x7, x8)
     char *fmt;
{
  recovery_with_line(lineno, fmt, x1, x2, x3, x4, x5, x6, x7, x8);
}



void
recovery_noline(fmt, x1, x2, x3, x4, x5, x6, x7, x8)
     char *fmt;
{
  if (infile != 0)
    {
      fprintf(stderr, "%s: ", infile);
      fprintf(stderr, fmt, x1, x2, x3, x4, x5, x6, x7, x8);
      putc('\n', stderr);
    }
  nerrors++;
}


void
recovery_with_line(lno, fmt, x1, x2, x3, x4, x5, x6, x7, x8)
     int lno;
     char *fmt;
{
  if (infile != 0)
    {
      fprintf(stderr, "%s:%d: ", infile, lno);
      fprintf(stderr, fmt, x1, x2, x3, x4, x5, x6, x7, x8);
      putc('\n', stderr);
    }
  nerrors++;
}


void
clear_recovery()
{
  nerrors = 0;
}


void
barrier_recovery()
{
  if (nerrors > 0)
    {
      if (infile != 0)
	fprintf(stderr, "Rie detects %d errors while parsing %s.\n", nerrors, infile);
      done(1);
    }
}


void
internal(s)
     char *s;
{
  fprintf(stderr, "<< Rie internal error. %s >>\n", s);
  fprintf(stderr, "<< Please let me know how situation you use.>>\n");
  done(1);
}


