/* Token-reader for Bison's input parser,
   Copyright (C) 1984, 1986, 1989 Free Software Foundation, Inc.

This file is part of Bison, the GNU Compiler Compiler.

Bison is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

Bison is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Bison; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */


/* 
   lex() is the entry point.  It is called from reader.c.
   It returns one of the token-type codes defined in lex.h.
   When an identifier is seen, the code IDENTIFIER is returned
   and the name is looked up in the symbol table using symtab.c;
   symval is set to a pointer to the entry found.  */

#include <stdio.h>
#include <ctype.h>
#include "system.h"
#include "files.h"
#include "symtab.h"
#include "lex.h"
#include "new.h"


extern int lineno;
extern int translations;

int parse_percent_token();
int parse_percent_token_for_semrule();

extern void fatals();
extern void fatal();

/* Buffer for storing the current token.  */
char *token_buffer;

/* Allocated size of token_buffer, not including space for terminator.  */
static int maxtoken;

bucket *symval;
int numval;

static int unlexed;		/* these two describe a token to be reread */
static bucket *unlexed_symval;	/* by the next call to lex */

static int unlexed_for_semrule; /* same as 'unlexed' but used in */
				/* semantic rule */

void
init_lex()
{
  maxtoken = 100;
  token_buffer = NEW2 (maxtoken + 1, char);
  unlexed = -1;
}


static char *
grow_token_buffer (p)
     char *p;
{
  int offset = p - token_buffer;
  maxtoken *= 2;
  token_buffer = (char *) xrealloc(token_buffer, maxtoken + 1);
  return token_buffer + offset;
}


int
skip_white_space()
{
  register int c;
  register int inside;

  c = getc(finput);

  for (;;)
    {
      int cplus_comment;

      switch (c)
	{
	case '/':
	  c = getc(finput);
	  if (c != '*' && c != '/')
#ifdef NOT_RIE
	    fatals("unexpected `/%c' found",c);
#endif
	    return ('/');
	  cplus_comment = (c == '/');

	  c = getc(finput);

	  inside = 1;
	  while (inside)
	    {
	      if (!cplus_comment && c == '*')
		{
		  while (c == '*')
		    c = getc(finput);

		  if (c == '/')
		    {
		      inside = 0;
		      c = getc(finput);
		    }
		}
	      else if (c == '\n')
		{
		  lineno++;
		  if (cplus_comment)
		    inside = 0;
		  c = getc(finput);
		}
	      else if (c == EOF)
		fatal("unterminated comment");
	      else
		c = getc(finput);
	    }

	  break;

	case '\n':
	  lineno++;

	case ' ':
	case '\t':
	case '\f':
	  c = getc(finput);
	  break;

	default:
	  return (c);
	}
    }
}


void
unlex(token)
int token;
{
  register char *cp;
  
#ifdef DEBUG1
      printf("unlex() : %d\n", token);
#endif

  switch (token)
    {
    case TWO_PERCENTS:
    case PERCENT_LEFT_CURLY:
    case TOKEN:
    case NTERM:
    case TYPE:	/* never used  */
    case GUARD:	/* never used  */
    case UNION:	/* never used  */
    case EXPECT:
    case START:
    case LEFT:
    case RIGHT:
    case NONASSOC:
    case SEMANTIC_PARSER:	/* never used  */
    case PURE_PARSER:		/* never used  */
    case PREC:
    case EQUIV:
      /* unput token strings  */
      cp = token_buffer;
      while (*cp)
	cp++;
      --cp;
      while (cp >= token_buffer)
	ungetc(*cp--, finput);
      ungetc('%', finput);
      unlexed = -1;
      return;
    }

  unlexed = token;
  unlexed_symval = symval;
}



int
lex()
{
  register int c;
  register char *p;

  if (unlexed >= 0)
    {
      symval = unlexed_symval;
      c = unlexed;
      unlexed = -1;
      return (c);
    }

  c = skip_white_space();

  switch (c)
    {
    case EOF:
      return (ENDFILE);

    case 'A':  case 'B':  case 'C':  case 'D':  case 'E':
    case 'F':  case 'G':  case 'H':  case 'I':  case 'J':
    case 'K':  case 'L':  case 'M':  case 'N':  case 'O':
    case 'P':  case 'Q':  case 'R':  case 'S':  case 'T':
    case 'U':  case 'V':  case 'W':  case 'X':  case 'Y':
    case 'Z':
    case 'a':  case 'b':  case 'c':  case 'd':  case 'e':
    case 'f':  case 'g':  case 'h':  case 'i':  case 'j':
    case 'k':  case 'l':  case 'm':  case 'n':  case 'o':
    case 'p':  case 'q':  case 'r':  case 's':  case 't':
    case 'u':  case 'v':  case 'w':  case 'x':  case 'y':
    case 'z':
    case '.':  case '_':
      p = token_buffer;
      while (isalnum(c) || c == '_' || c == '.')
	{
	  if (p == token_buffer + maxtoken)
	    p = grow_token_buffer(p);

	  *p++ = c;
	  c = getc(finput);
	}

      *p = 0;
      ungetc(c, finput);
      symval = getsym(token_buffer);
      return (IDENTIFIER);

    case '0':  case '1':  case '2':  case '3':  case '4':
    case '5':  case '6':  case '7':  case '8':  case '9':
      {
	numval = 0;

	while (isdigit(c))
	  {
	    numval = numval*10 + c - '0';
	    c = getc(finput);
	  }
	ungetc(c, finput);
	return (NUMBER);
      }

    case '\'':
      translations = -1;

      /* parse the literal token and compute character code in  code  */

      c = getc(finput);
      {
	register int code = 0;

	if (c == '\\')
	  {
	    c = getc(finput);

	    if (c <= '7' && c >= '0')
	      {
		while (c <= '7' && c >= '0')
		  {
		    code = (code * 8) + (c - '0');
		    c = getc(finput);
		    if (code >= 256 || code < 0)
		      fatals("malformatted literal token `\\%03o'", code);
		  }
	      }
	    else
	      {
		if (c == 't')
		  code = '\t';
		else if (c == 'n')
		  code = '\n';
		else if (c == 'a')
		  code = '\007';
		else if (c == 'r')
		  code = '\r';
		else if (c == 'f')
		  code = '\f';
		else if (c == 'b')
		  code = '\b';
		else if (c == 'v')
		  code = 013;
		else if (c == 'x')
		  {
		    c = getc(finput);
		    while ((c <= '9' && c >= '0')
			   || (c >= 'a' && c <= 'z')
			   || (c >= 'A' && c <= 'Z'))
		      {
			code *= 16;
			if (c <= '9' && c >= '0')
			  code += c - '0';
			else if (c >= 'a' && c <= 'z')
			  code += c - 'a' + 10;
			else if (c >= 'A' && c <= 'Z')
			  code += c - 'A' + 10;
			if (code >= 256 || code<0)/* JF this said if(c>=128) */
			  fatals("malformatted literal token `\\x%x'",code);
			c = getc(finput);
		      }
		    ungetc(c, finput);
		  }
		else if (c == '\\')
		  code = '\\';
		else if (c == '\'')
		  code = '\'';
		else if (c == '\"')	/* JF this is a good idea */
		  code = '\"';
		else
		  {
		    if (c >= 040 && c <= 0177)
		      fatals ("unknown escape sequence `\\%c'", c);
		    else
		      fatals ("unknown escape sequence: `\\' followed by char code 0x%x", c);
		  }

		c = getc(finput);
	      }
	  }
	else
	  {
	    code = c;
	    c = getc(finput);
	  }
	if (c != '\'')
	  fatal("multicharacter literal tokens not supported");

	/* now fill token_buffer with the canonical name for this character
	   as a literal token.  Do not use what the user typed,
	   so that '\012' and '\n' can be interchangeable.  */

	p = token_buffer;
	*p++ = '\'';
	if (code == '\\')
	  {
	    *p++ = '\\';
	    *p++ = '\\';
	  }
	else if (code == '\'')
	  {
	    *p++ = '\\';
	    *p++ = '\'';
	  }
	else if (code >= 040 && code != 0177)
	  *p++ = code;
	else if (code == '\t')
	  {
	    *p++ = '\\';
	    *p++ = 't';
	  }
	else if (code == '\n')
	  {
	    *p++ = '\\';
	    *p++ = 'n';
	  }
	else if (code == '\r')
	  {
	    *p++ = '\\';
	    *p++ = 'r';
	  }
	else if (code == '\v')
	  {
	    *p++ = '\\';
	    *p++ = 'v';
	  }
	else if (code == '\b')
	  {
	    *p++ = '\\';
	    *p++ = 'b';
	  }
	else if (code == '\f')
	  {
	    *p++ = '\\';
	    *p++ = 'f';
	  }
        else
	  {
	    *p++ = code / 0100 + '0';
	    *p++ = ((code / 010) & 07) + '0';
	    *p++ = (code & 07) + '0';
	  }
	*p++ = '\'';
	*p = 0;
	symval = getsym(token_buffer);
	symval->class = STOKEN;
	if (! symval->user_token_number)
	  symval->user_token_number = code;
	return (IDENTIFIER);
      }

    case ',':
      return (COMMA);

    case ':':
      return (COLON);

    case ';':
      return (SEMICOLON);

    case '|':
      return (BAR);

    case '{':
      return (LEFT_CURLY);

    case '=':
      do
	{
	  c = getc(finput);
	  if (c == '\n') lineno++;
	}
      while(c==' ' || c=='\n' || c=='\t');

      if (c == '{')
      	return(LEFT_CURLY);
      else
	{
	  ungetc(c, finput);
	  return(ILLEGAL);
	}

    case '<':
      p = token_buffer;
      c = getc(finput);
      while (c != '>')
	{
	  if (c == '\n' || c == EOF)
	    fatal("unterminated type name");

	  if (p == token_buffer + maxtoken)
	    p = grow_token_buffer(p);

	  *p++ = c;
	  c = getc(finput);
	}
      *p = 0;
      return (TYPENAME);
	    

    case '%':
      return (parse_percent_token());

    default:
      return (ILLEGAL);
    }
}



/* get token for attribute definition.  */
int
lex_for_attribute()
{
  register int c;
  register char *p;

  if (unlexed >= 0)
    {
      symval = unlexed_symval;
      c = unlexed;
      unlexed = -1;
#ifdef DEBUG1
      printf("lex_for_attribute() : unlexed token %d\n", c);
#endif
      return (c);
    }

  c = skip_white_space();

  if (isalpha(c) || c == '_')
    {
      p = token_buffer;

      while (isalnum(c) || c == '_')
	{
	  if (p == token_buffer + maxtoken)
	    p = grow_token_buffer(p);
	      
	  *p++ = c;
	  c = getc(finput);
	}

      *p = '\0';
      ungetc(c, finput);

#ifdef DEBUG1
      printf("lex_for_attribute() : reads IDENTIFIER \"%s\"\n", token_buffer);
#endif
      if (strcmp(token_buffer, "inh") == 0)
	return (INHT);
      else if (strcmp(token_buffer, "synt") == 0)
	return (SYNT);
      else
	return (IDENTIFIER);
    }

  /* 'c' is no more a part of identifier */
  
  switch (c)
    {
    case EOF:
#ifdef DEBUG1
	printf("lex_for_attribute() : reads EOF\n");
#endif
      return (ENDFILE);

    case ',':
#ifdef DEBUG1
	printf("lex_for_attribute() : reads COMMA\n");
#endif
      return (COMMA);

    case ':':
#ifdef DEBUG1
	printf("lex_for_attribute() : reads COLON\n");
#endif
      return (COLON);
	
    case ';':
#ifdef DEBUG1
	printf("lex_for_attribute() : reads SEMICOLON\n");
#endif
      return (SEMICOLON);

    case '*':
#ifdef DEBUG1
	printf("lex_for_attribute() : reads ASTERISK\n");
#endif
      return (ASTERISK);

    case '.':
#ifdef DEBUG1
	printf("lex_for_attribute() : reads PERIOD\n");
#endif
      return (PERIOD);
      
    case '%':
#ifdef DEBUG1
	printf("lex_for_attribute() : reads PERCENT\n");
#endif
      return (parse_percent_token());

    default:
#ifdef DEBUG1
	printf("lex_for_attribute() : reads ILLEGAL\n");
#endif
      return (ILLEGAL);
    }
}


/* parse a token which starts with %.  Assumes the % has already been read and discarded.  */

int
parse_percent_token ()
{
  register int c;
  register char *p;

  p = token_buffer;
  c = getc(finput);

  *p++ = c;
  *p = '\0';
  
  switch (c)
    {
    case '%':
      return (TWO_PERCENTS);

    case '{':
      return (PERCENT_LEFT_CURLY);

    case '<':
      return (LEFT);

    case '>':
      return (RIGHT);

    case '2':
      return (NONASSOC);

    case '0':
      return (TOKEN);

    case '=':
      return (PREC);
    }
  if (!isalpha(c))
    return (ILLEGAL);

  p = token_buffer;

  while (isalpha(c) || c == '_')
    {
      if (p == token_buffer + maxtoken)
	p = grow_token_buffer(p);

      *p++ = c;
      c = getc(finput);
    }

  ungetc(c, finput);

  *p = 0;

  if (strcmp(token_buffer, "token") == 0
      ||
      strcmp(token_buffer, "terminal") == 0
      ||
      strcmp(token_buffer, "term") == 0)
    return (TOKEN);
  else if (strcmp(token_buffer, "nterm") == 0)
    return (NTERM);
  else if (strcmp(token_buffer, "nonterm") == 0)
    return (NTERM);
  else if (strcmp(token_buffer, "type") == 0)
    return (TYPE);
  else if (strcmp(token_buffer, "guard") == 0)
    return (GUARD);
  else if (strcmp(token_buffer, "union") == 0)
    return (UNION);
  else if (strcmp(token_buffer, "expect") == 0)
    return (EXPECT);
  else if (strcmp(token_buffer, "start") == 0)
    return (START);
  else if (strcmp(token_buffer, "left") == 0)
    return (LEFT);
  else if (strcmp(token_buffer, "right") == 0)
    return (RIGHT);
  else if (strcmp(token_buffer, "nonassoc") == 0
	   ||
	   strcmp(token_buffer, "binary") == 0)
    return (NONASSOC);
  else if (strcmp(token_buffer, "semantic_parser") == 0)
    return (SEMANTIC_PARSER);
  else if (strcmp(token_buffer, "pure_parser") == 0)
    return (PURE_PARSER);
  else if (strcmp(token_buffer, "prec") == 0)
    return (PREC);
  else if (strcmp(token_buffer, "equiv") == 0)
    return (EQUIV);
  else return (ILLEGAL);
}



/* initialize lexical analyzer used inside semantic acrion. */
void
init_lex_for_semrule()
{
  unlexed_for_semrule = -1;
}



/* get one token inside semantic action.  */
int
lex_for_semrule()
{
  
  register int c;
  register char *p;

  if (unlexed_for_semrule >= 0)
    {
      c = unlexed_for_semrule;
      unlexed_for_semrule = -1;
      return (c);
    }
  
  c = skip_white_space();

  p = token_buffer;
  *p++ = c;
  *p = '\0';

  if (isalpha(c) || c == '_')
    {
      p--;

      while (isalnum(c) || c == '_')
	{
	  if (p == token_buffer + maxtoken)
	    p = grow_token_buffer(p);

	  *p++ = c;
	  c = getc(finput);
	}

      *p = '\0';
      ungetc(c, finput);

#ifdef DEBUG1
      printf("lex_for_semrule() : return IDENTIFIER \"%s\"\n",token_buffer);
#endif
      return (IDENTIFIER);
     }

  /* 'c' is no more a part of identifier.  */

  switch (c)
    {
    case EOF:
#ifdef DEBUG1
      printf("lex_for_semrule() : return ENDFILE\n");
#endif
      return (ENDFILE);

    case '[':
      {
	/* perhaps an index */
	register unsigned level;
	
	/* peek one character */
	c = getc(finput);
	ungetc(c, finput);

	if (isspace(c))
	  *p++ = ' ';

	c = skip_white_space();

	if (isdigit(c))
	  {
	    numval = 0;
	    while(isdigit(c))
	      {
		numval = numval*10 + c - '0';
		if (p == token_buffer + maxtoken)
		  p = grow_token_buffer(p);
		*p++ = c;
		c = getc(finput);
	      }
	    ungetc(c, finput);
	  
	    if (isspace(c))
	      {
		if (p == token_buffer + maxtoken)
		  p = grow_token_buffer(p);
		*p++ = ' ';
	      }

	    c = skip_white_space();
	    if (c == ']')
	      {
		if (p == token_buffer + maxtoken)
		  p = grow_token_buffer(p);
		*p++ = ']';
		*p = '\0';
#ifdef DEBUG1
		printf("lex_for_semrule() : return INDEX %d\n",numval);
#endif
		return (INDEX);
	      }
	  }
      
	/*
	 * This doesn't works correctly on strict C grammer.
	 * o no checking string constants in [].
	 * o need white spaces before comments.
	 * o "a+" and "a +" are different. but "a +", "a  +", "a   +"
	 *    are same.
	 */	
	level = 1;
	while(level != 0)
	  {
	    /* peek one character  */
	    c = getc(finput);
	    ungetc(c, finput);
	  
	    if (isspace(c))
	      {
		if (p == token_buffer + maxtoken)
		  p = grow_token_buffer(p);
		*p++ = ' ';
	      }

	    c = skip_white_space();
	    if (p == token_buffer + maxtoken)
	      p = grow_token_buffer(p);
	    *p++ = c;

	    if (c == ']')
	      level--;
	    else if (c == '[')
	      level++;
	    else if (c == EOF)
	      fatal("unmatched '['");
	  }
	  
	/* remove last one space  */
	if (p > token_buffer && *(p-1) == ' ')
	  p--;
	
	*p = '\0';
#ifdef DEBUG1
	printf("lex_for_semrule() : return CODES \"%s\"\n",token_buffer);
#endif
	return (CODES);
      }

    case '\'':
    case '"':
      {
	register char match;

	match = c;
	c = getc(finput);

	while (c != match)
	  {
	    if (c == EOF || c == '\n')
	      fatal("unterminated string");

	    if (p == token_buffer + maxtoken)
	      p = grow_token_buffer(p);
	    *p++ = c;

	    if (c == '\\')
	      {
		c = getc(finput);
		if (c == EOF)
		  fatal("unterminated string");
		if (p == token_buffer + maxtoken)
		  p = grow_token_buffer(p);
		*p++ = c;
		if (c == '\n')
		  lineno++;
	      }

	    c = getc(finput);
	  }
	if (p == token_buffer + maxtoken)
	  p = grow_token_buffer(p);
	*p++ = c;
	*p = '\0';

#ifdef DEBUG1
      printf("lex_for_semrule() : return STRING %s\n", token_buffer);
#endif
	return (STRING);
      }
	
    case '(':
#ifdef DEBUG1
      printf("lex_for_semrule() : return LEFT_PAREN\n");
#endif
      return (LEFT_PAREN);

    case ')':
#ifdef DEBUG1
      printf("lex_for_semrule() : return RIGHT_PAREN\n");
#endif
      return (RIGHT_PAREN);
      
    case '{':
#ifdef DEBUG1
      printf("lex_for_semrule() : return LEFT_CURLY\n");
#endif
      return (LEFT_CURLY);

    case '}':
#ifdef DEBUG1
      printf("lex_for_semrule() : return RIGHT_CURLY\n");
#endif
      return (RIGHT_CURLY);
      
    case '=':
#ifdef DEBUG1
      printf("lex_for_semrule() : return ASSIGN\n");
#endif
      return (ASSIGN);

    case '.':
#ifdef DEBUG1
      printf("lex_for_semrule() : return PERIOD\n");
#endif
      return (PERIOD);

    case ';':
#ifdef DEBUG1
      printf("lex_for_semrule() : return SEMICOLON\n");
#endif
      return (SEMICOLON);

    case ',':
#ifdef DEBUG1
      printf("lex_for_semrule() : return COMMA\n");
#endif
      return (COMMA);

    case '%':
      return (parse_percent_token_for_semrule());

    default:
      /* code */
      p--;
      /* THINK MORE !! A STATEMENT BELLOWS IS NOT GOOD !! */
      while(! (isalpha (c) || c == '_' || c == ';' || c == ',' || c == '}'
	       || c == '(' || c == ')' || c == '"' || c == '\''))
	{
	  if (p == token_buffer + maxtoken)
	    p = grow_token_buffer(p);
	  *p++ = c;

	  /* peek one character  */
	  c = getc(finput);
	  ungetc(c, finput);

	  if (isspace(c))
	    {
	      if (p == token_buffer + maxtoken)
		p = grow_token_buffer(p);
	      *p++ = ' ';
	    }
	  c = skip_white_space();
	}
      ungetc(c, finput);

      /* remove last one space  */
      if (p > token_buffer && *(p-1) == ' ')
	p--;
      
      *p = '\0';
#ifdef DEBUG1
      printf("lex_for_semrule() : return CODES \"%s\"\n", token_buffer);
#endif
      return (CODES);
    }
}


/* parse reserved token begging with % inside semantic rule.  */
int
parse_percent_token_for_semrule()
{
  register int c;
  register char *p;

  c = getc(finput);

  if (! isalpha(c))
    {
      ungetc(c, finput);
#ifdef DEBUG1
      printf("parse_percert_token_for_semrule() : return CODES \"%s\"\n"
	     ,token_buffer);
#endif
      return (CODES);
    }

  p = token_buffer;

  while (isalpha(c))
    {
      if (p == token_buffer + maxtoken)
	p = grow_token_buffer(p);
      *p++ = c;
      c = getc(finput);
    }
  ungetc(c, finput);
  *p = '\0';

  if (strcmp(token_buffer, "local") == 0)
#ifdef DEBUG1
    {
      printf("parse_percert_token_for_semrule() : return LOCALS\n");
#endif
    return (LOCALS);
#ifdef DEBUG1
    }
#endif
  else if (strcmp(token_buffer, "transfer") == 0)
#ifdef DEBUG1
    {
      printf("parse_percert_token_for_semrule() : return TRANSFER\n");
#endif
    return (TRANSFER);
#ifdef DEBUG1
    }
#endif
  else if (strcmp(token_buffer, "thread") == 0)
#ifdef DEBUG1
    {
      printf("parse_percert_token_for_semrule() : return THREAD\n");
#endif
    return (THREAD);
#ifdef DEBUG1
    }
#endif
  else if (strcmp(token_buffer, "except") == 0)
#ifdef DEBUG1
    {
      printf("parse_percert_token_for_semrule() : return EXCEPT\n");
#endif
    return (EXCEPT);
#ifdef DEBUG1
    }
#endif
  else if (strcmp(token_buffer, "condition") == 0)
#ifdef DEBUG1
    {
      printf("parse_percert_token_for_semrule() : return CONDITION\n");
#endif
    return (CONDITION);
#ifdef DEBUG1
    }
#endif
  else if (strcmp(token_buffer, "message") == 0)
#ifdef DEBUG1
    {
      printf("parse_percert_token_for_semrule() : return MESSAGE\n");
#endif
    return (MESSAGE);
#ifdef DEBUG1
    }
#endif
  else if (strcmp(token_buffer, "action") == 0)
#ifdef DEBUG1
    {
      printf("parse_percert_token_for_semrule() : return ACTION\n");
#endif
    return (ACTION);
#ifdef DEBUG1
    }
#endif
  else if (strcmp(token_buffer, "debug") == 0)
#ifdef DEBUG1
    {
      printf("parse_percert_token_for_semrule() : return DEBUGS\n");
#endif
    return (DEBUGS);
#ifdef DEBUG1
    }
#endif
  else
#ifdef DEBUG1
    {
      printf("parse_percert_token_for_semrule() : return ILLEGAL \"%s\"\n"
	     , token_buffer);
#endif
    return (ILLEGAL);
#ifdef DEBUG1
    }
#endif
}



/* unput token */

void
unlex_for_semrule(token)
int token;
{
  unlexed_for_semrule = token;
}


/* unget token. return the stoken string to input.  */
void
send_back_token()
{
  register char *cp;
  
  if (unlexed_for_semrule < 0)
    return;

  /* send back  */
  cp = token_buffer;
  while (*cp)
    cp++;
  --cp;
  while (cp >= token_buffer)
    ungetc(*cp--, finput);

  unlexed_for_semrule = -1;
}


/* get one token inside C block.
   return value is one of
   IDENTIFIER, INDEX, CODES, LEFT_CURLY, RIGHT_CURLY, PERIOD or EOF.  */

int
lex_for_C_block()
{
  int c, cplus_comment, inside;
  char *p;

  if (unlexed_for_semrule >= 0)
    send_back_token();
  
  c = getc(finput);

  p = token_buffer;
  *p++ = c;
  *p = '\0';

  if (isalpha(c) || c == '_')
    {
      p--;

      while (isalnum(c) || c == '_')
	{
	  if (p == token_buffer + maxtoken)
	    p = grow_token_buffer(p);

	  *p++ = c;
	  c = getc(finput);
	}

      *p = '\0';
      ungetc(c, finput);

#ifdef DEBUG1
      printf("lex_for_C_block() : return IDENTIFIER \"%s\"\n",token_buffer);
#endif
      return (IDENTIFIER);
    }

  /* 'c' is no more a part of identifier.  */

  switch (c)
    {
    case EOF:
#ifdef DEBUG1
      printf("lex_for_C_block() : return ENDFILE\n");
#endif
      return (ENDFILE);

    case '[':
      {
	/* perhaps an index.
	   parse the regular expression '\[[ \t\n]*[0-9]+[ \t\n]*\]'. */
	unsigned level;
	
	c = getc(finput);

	/* copy white spaces preceding a number  */
	for (; isspace(c); c=getc(finput))
	  {
	    if (c == '\n')
	      lineno++;
	    if (p == token_buffer + maxtoken)
	      p = grow_token_buffer(p);
	    *p++ = c;
	  }
	
	/* is this an index allowed to be used for attributes ?  */
	if (isdigit(c))
	  {
	    numval = 0;
	    while(isdigit(c))
	      {
		numval = numval*10 + c - '0';
		if (p == token_buffer + maxtoken)
		  p = grow_token_buffer(p);
		*p++ = c;
		c = getc(finput);
	      }
	  
	    /* copy white spaces following a number  */
	    for (; isspace(c); c=getc(finput))
	      {
		if (c == '\n')
		  lineno++;
		if (p == token_buffer + maxtoken)
		  p = grow_token_buffer(p);
		*p++ = c;
	      }

	    if (c == ']')
	      {
		/* yes, this is really an index for attribute.  */
		if (p == token_buffer + maxtoken)
		  p = grow_token_buffer(p);
		*p++ = ']';
		*p = '\0';
#ifdef DEBUG1
		printf("lex_for_C_block() : return INDEX %d\n",numval);
#endif
		return (INDEX);
	      }
	    /* otherwise, treat as a code  */
	  }
		   
	/*
	 * This doesn't works correctly on strict C grammer.
	 * no checking string constants or comments in [].
	 */	
	level = 1;
	for (; level != 0; c=getc(finput))
	  {
	    switch (c)
	      {
	      case ']':
		--level;
		break;

	      case '[':
		++level;
		break;

	      case '\n':
		lineno++;
		break;
		
	      case EOF:
		fatal("unmatched '['");

	      default:
		/* do nothing */
		;
	      }

	    if (p == token_buffer + maxtoken)
	      p = grow_token_buffer(p);
	    *p++ = c;
	  }
	ungetc(c,finput);
	*p = '\0';
#ifdef DEBUG1
	printf("lex_for_C_block() : return CODES \"%s\"\n",token_buffer);
#endif
	return (CODES);
      }

    case '\'':
    case '"':
      {
	char match;

	match = c;
	c = getc(finput);

	while (c != match)
	  {
	    if (c == EOF || c == '\n')
	      fatal("unterminated string");

	    if (p == token_buffer + maxtoken)
	      p = grow_token_buffer(p);
	    *p++ = c;

	    if (c == '\\')
	      {
		c = getc(finput);
		if (c == EOF)
		  fatal("unterminated string");
		if (p == token_buffer + maxtoken)
		  p = grow_token_buffer(p);
		*p++ = c;
		if (c == '\n')
		  lineno++;
	      }

	    c = getc(finput);
	  }
	if (p == token_buffer + maxtoken)
	  p = grow_token_buffer(p);
	*p++ = c;
	*p = '\0';

#ifdef DEBUG1
	printf("lex_for_C_block() : return CODES(STRING) %s\n", token_buffer);
#endif
	return (CODES);
      }
	
    case '/':
      c = getc(finput);
      if (c != '*' && c != '/')
	{
	  /* not a comment */
	  ungetc(c, finput);
	  *p = '\0';
#ifdef DEBUG1
	printf("lex_for_C_block() : return CODES \"%s\"\n", token_buffer);
#endif
	  return (CODES);
	}

      cplus_comment = (c == '/');
      *p++ = c;			/* not check boundary */
      c = getc(finput);
      
      inside = 1;
      for (; inside; c=getc(finput))
	{
	  if (p == token_buffer + maxtoken)
	    p = grow_token_buffer(p);
	  *p++ = c;
	  
	  switch (c)
	    {
	    case '*':
	      if (!cplus_comment)
		{
		  for (; c == '*'; c=getc(finput))
		    {
		      if (p == token_buffer + maxtoken)
			p = grow_token_buffer(p);
		      *p++ = c;
		    }

		  if (c == '/')
		    {
		      if (p == token_buffer + maxtoken)
			p = grow_token_buffer(p);
		      *p++ = c;
		      inside = 0;	/* for exit for loop */
		    }
		  else
		    ungetc(c, finput);
		}
	      break;
	      
	    case '\n':
	      lineno++;
	      if (cplus_comment)
		inside = 0;	/* for exit for loop */
	      break;

	    case EOF:
	      fatal("unterminated comment");

	    default:
	      /* do nothing */
	      break;
	    }
	}	/* for */
      ungetc(c, finput);
      *p = '\0';
#ifdef DEBUG1
	printf("lex_for_C_block() : return CODES(COMMENT) %s\n", token_buffer);
#endif
      return (CODES);

    case '{':
#ifdef DEBUG1
      printf("lex_for_C_block() : return LEFT_CURLY\n");
#endif
      return (LEFT_CURLY);

    case '}':
#ifdef DEBUG1
      printf("lex_for_C_block() : return RIGHT_CURLY\n");
#endif
      return (RIGHT_CURLY);
      
    case '.':
#ifdef DEBUG1
      printf("lex_for_C_block() : return PERIOD\n");
#endif
      return (PERIOD);

    case '\n':
      lineno++;
      /* through */
      
    default:
      /* code */
      c = getc(finput);
      /* THINK MORE !! A STATEMENT BELLOWS IS NOT GOOD !! */
      for (; ! (isalpha (c) || c == '_' || c == '{' || c == '}'
		|| c == '/' || c == '"' || c == '\''); c=getc(finput))
	{
	  if (c == '\n')
	    lineno++;
	  if (p == token_buffer + maxtoken)
	    p = grow_token_buffer(p);
	  *p++ = c;
	}
      ungetc(c, finput);
      *p = '\0';
#ifdef DEBUG1
      printf("lex_for_C_block() : return CODES \"%s\"\n", token_buffer);
#endif
      return (CODES);
    }
}


/* the function 'unlex_for_C_block()' is identical to 'unlex_for_semrule()' */

void
(*unlex_for_C_block)() = unlex_for_semrule;


/* discard tokens in the declaration section for error recovery.  */

void
discard_decl()
{
  register int token;

  for (;;)
    {
      if (ungetc(skip_white_space(), finput) == '%')
	return;

      token = lex_for_attribute();

      switch (token)
	{
	case COMMA:
	case SEMICOLON:
	  unlex(token);
	  return;
	}
    }
}



/* discard tokens in the semantic rule for error recovery..  */

void
discard_semrule()
{
  register int token;

  for(;;)
    {
      token = lex_for_semrule();

      switch (token)
	{
	case SEMICOLON:
	case RIGHT_CURLY:
	case LOCALS:
	case TRANSFER:
	case THREAD:
	case EXCEPT:
	case CONDITION:
	case MESSAGE:
	case ACTION:
	case DEBUGS:
	  unlex_for_semrule(token);
	  return;
	}
    }
}
