/* attribute output functions
   Copyright (C) 1991-1993 Masataka Sassa, Kazuhiro Kuroishi, Teruhisa Hirai.
   Copyright (C)      1994,1995                              Yoshiki Ohshima.
     
This file is part of Rie, A Compiler Compiler Based on ECLR-attributed Grammar.

Rie is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

Rie is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Bison; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

Bug reports, fixes or ideas are welcomed.
		Address: Yoshiki Ohshima
			 Tokyo Institute of Technology,
			 Department of Information Science,
			 Ookayama, Meguro-ku, Tokyo, 152 Japan.

		E-mail: rie-comments@is.titech.ac.jp			*/

#include <stdio.h>
#include "attribute.h"
#include "files.h"
#include "gram.h"
#include "new.h"
#include "symtab.h"
#include "system.h"

extern char *infile;			/* files.c  */
extern int lineno;			/* reader.c  */
extern int nolinesflag;			/* getargs.c  */
extern semantic_rule **inh_table;	/* attribute.c  */
extern semantic_rule **synt_table;	/* attribute.c  */
extern semantic_rule **local_table;	/* attribute.c  */
extern semantic_condition **condition_table;	/* attribute.c  */
extern int nsynthesizes;		/* attribute.c  */
extern int nlocal_classes;		/* attribute.c  */
extern int nEC;				/* attribute.c  */
extern short *rrhsend;			/* attribute.c  */

#if (defined(__STDC__) || defined(__GNUC__))

extern void internal(char *);			/* error.c  */
extern pure_list *copy_pure_lists(pure_list *);	/* pure.c  */
extern void free_pure_lists(pure_list *);	/* pure.c  */

#else	/* not ANSI C nor GNU C  */

extern void internal();			/* error.c  */
extern pure_list *copy_pure_lists();	/* pure.c  */
extern void free_pure_lists();		/* pure.c  */

#endif	/* not ANSI C nor GNU C  */

typedef
  struct string_list 
    {
      struct string_list *link;
      char *tag;
    }
  string_list;

typedef
  struct synt_pack	
    {
      struct synt_pack *link;
      string_list *string_list;
      int npures;
      pure_list *pure_list;
    }
  synt_pack;


#if (defined(__STDC__) || defined(__GNUC__))

void outputsem(void);
void output_synt_type(void);
void output_inherit_header(void);
void output_inherit_trailer(void);
void output_inherit_pre(int);
void output_inherit_post(void);
void output_inherit(semantic_rule **);
void output_local(semantic_rule_list *);
void output_line_pragma(void);
void output_semexps(FILE *, semantic_expression *, int);
void output_attribute_type(void);

static void pack_synt_type(void);
static void free_synt_type(void);
static void output_inh_type(void);
static void pack_inh_type(void);
static void output_local_type(void);
static void pack_local_type(void);
static void output_synthesize(void);
static void output_attribute_defines(void);
static void output_semrule(FILE *, semantic_rule *);
static void output_semrule_plain(FILE *, semantic_rule *);
static void output_lhs_attribute(FILE *, attribute *);
       void output_semantic_condition(FILE *, semantic_condition *);

#else	/* not ANSI C nor GNU C  */

void outputsem();
void output_synt_type();
void output_inherit_header();
void output_inherit_trailer();
void output_inherit_pre();
void output_inherit_post();
void output_inherit();
void output_local();
void output_line_pragma();
void output_semexps();
void output_attribute_type();

static void pack_synt_type();
static void free_synt_type();
static void output_inh_type();
static void pack_inh_type();
static void output_local_type();
static void pack_local_type();
static void output_synthesize();
static void output_attribute_defines();
static void output_semrule();
static void output_semrule_plain();
static void output_lhs_attribute();
       void output_semantic_condition();

#endif	/* not ANSI C nor GNU C  */

static synt_pack *synt_type;
static char **inh_type;
static attribute_list **local_type;

/* output type declarations of attribute and semantic rules.

   some of output functions, which output type declaration of attributes and
   semantic rules, are called in other functions.
   output of type declarations of synthesized attribute is at reader().
   output of semantic rules of inherited and local attribute is at eclr().  */

void
outputsem()
{
    /* type declarations of inherited attributes.  */
    if (nEC > 0)
	output_inh_type();

    /* type declarations of local attributes.  */
    output_local_type();
    
    /* semantic rules of synthesized attributes.  */
    output_synthesize();

    /* some macros for parser.  */
    output_attribute_defines();
}



/* output types of synthesized attribute.  */

void
output_synt_type()
{
  register synt_pack *types;
  register pure_list *p1;
  register string_list *sp;
  register int count;
  
  if (nsynthesizes == 0)
    return;

  synt_type = NULL;
  pack_synt_type();

  if (synt_type == NULL)
    return;

  /* all types are packed. print all  */
  fprintf(ftype, "typedef\n  union\n    {\n");
  if (fdefines)
    fprintf(fdefines, "typedef\n  union\n    {\n");
  
  types = synt_type;
  while (types)
    {
      fprintf(ftype, "      struct {\n");
      if (fdefines)
	fprintf(fdefines, "      struct {\n");
      
      p1 = types->pure_list;
      while (p1)
	{
	  fprintf(ftype, "\t%s\t%s;\n", p1->pure->type, p1->pure->name);
	  if (fdefines)
	    fprintf(fdefines, "\t%s\t%s;\n",
		    p1->pure->type, p1->pure->name);
	  p1 = p1->link;
	}
      fprintf(ftype, "      } rr%s", types->string_list->tag);
      if (fdefines)
	fprintf(fdefines, "      } rr%s", types->string_list->tag);

      count = 1;
      sp = types->string_list->link;
      while (sp)
	{
	  fprintf(ftype, ", rr%s", sp->tag);
	  if (fdefines)
	    fprintf(fdefines, ", rr%s", sp->tag);
	  if (++count > 4 && sp->link)
	    {
	      count = 0;
	      fprintf(ftype, "\n      ");
	      if (fdefines)
		fprintf(fdefines, "\n      ");
	    }
	  sp = sp->link;
	}
      fprintf(ftype, ";\n\n");
      if (fdefines)
	fprintf(fdefines, ";\n\n");

      types = types->link;
    }

  fprintf(ftype, "    }\n  rrsastype;\n\n");
  if (fdefines)
    {
      fprintf(fdefines, "    }\n  rrsastype;\n\n");
      fprintf(fdefines, "extern rrsastype rrlval;\n\n");
    }
  
  free_synt_type();
}



/* pack types of all synthesized attributes  */

static
void
pack_synt_type()
{
  register synt_pack *types;
  register pure_list *p1, *p2, *sub, *super, *purelist;
  register string_list *sp;
  register bucket *symbol;
  register int npures;
  
#ifdef DEBUG1
  printf("\n\nPacking types of all synthesized attribute...\n");
#endif
  
  /* skip "error" and "$illegal." tokens  */
  symbol = firstsymbol->next->next;
  
  for (; symbol; symbol = symbol->next)
    {
      /* for all symbol  */

      /* count the number of pures and enlist them. pures' class is
	 only SYNTHESIZE.  */
      p1 = purelist = copy_pure_lists(symbol->pure_list);
      p2 = NULL;		/* points previous list  */
      npures = 0;
      while (p1)
	{
	  if (p1->pure->class == SYNTHESIZE)
	    {
	      npures++;
	      p2 = p1;
	      p1 = p1->link;
	    }
	  /* not synthesized attribute.  skip.  */
	  else if (p2)
	    {
	      p2->link = p1->link;
	      FREE(p1);
	      p1 = p2->link;
	    }
	  else
	    {
	      purelist = p1->link;
	      FREE(p1);
	      p1 = purelist;
	    }
	}

      if (npures == 0)
	continue;
      
      /* check types with defined type pack list.  */
      types = synt_type;
      while (types)
	{
	  if (npures <= types->npures)
	    {
	      sub = purelist;
	      super = types->pure_list;
	    }
	  else
	    {
	      sub = types->pure_list;
	      super = purelist;
	    }
	  
	  p1 = sub;
	  while (p1)
	    {
	      p2 = super;
	      while (p2)
		{
		  /* IF YOU DON'T USE THE STRING MANAGER.
		     USE 'strcmp()' INSTEAD.  */
		  if (p1->pure->name == p2->pure->name
		      &&
		      p1->pure->type == p2->pure->type)
		    break;
		  p2 = p2->link;
		}

	      if (p2 == NULL)
		/* there is no same pure as 'p1->pure' in the list of 'super'.
		   'super' is not the superset of 'sub'.  */
		break;

	      p1 = p1->link;
	    }

	  if (p1 == NULL)
	    {
	      /* 'super' is the superset of 'sub'.  */
	      if (super == purelist)
		{
		  /* pures of 'types' are included that of 'symbol'.  */
		  free_pure_lists(types->pure_list);
		  types->npures = npures;
		  types->pure_list = purelist;
		}
	      else
		free_pure_lists(purelist);
	      
	      sp = NEW(string_list);
	      sp->tag = symbol->tag;
	      sp->link = types->string_list;
	      types->string_list = sp;

	      break;		/* while (types)  */
	    }
	  
	  types = types->link;
	}			/* while(types)  */
      
      if (types == NULL)
	{
	  /* the set of pures of 'purelist' is not the subset nor
	     superset of sets defined already.  */

	  /* create new type pack.  */
	  types = NEW(synt_pack);
	  sp = NEW(string_list);

	  sp->tag = symbol->tag;
	  sp->link = NULL;
	  types->string_list = sp;
	  types->npures = npures;
	  types->pure_list = purelist;
	  types->link = synt_type;
	  synt_type = types;
	}
    }	/* for (symbol)  */
}



/* free 'synt_type'.  */

static
void
free_synt_type()
{
  register synt_pack *tp, *nextt;
  register string_list *sp, *nexts;
  
  tp = synt_type;
  while (tp)
    {
      
      sp = tp->string_list;
      while (sp)
	{
	  nexts = sp->link;
	  FREE(sp);
	  sp = nexts;
	}

      free_pure_lists(tp->pure_list);
      
      nextt = tp->link;
      FREE(tp);
      tp = nextt;
    }
}



/* output types of all inherited attributes to 'ftype'.  */

static
void
output_inh_type()
{
  register int i;

  inh_type = NEW2(nEC, char *);
  pack_inh_type();

  fprintf(ftype, "typedef\n  struct\n    {\n");
  for (i=0; i<nEC; i++)
    {
      if (inh_type[i])
	fprintf(ftype, "      %s\tEC%d;\n",inh_type[i], i);
    }
  fprintf(ftype, "    }\n  rriastype;\n\n");

  FREE(inh_type);
}



/* pack types of all inherited attributes  */

static
void
pack_inh_type()
{
  register int nec;
  register semantic_rule **tp, **tpend, *rp;
  register short ec;
  
  nec = 0;
  
  tp = inh_table;
  tpend = inh_table + nitems;
  while (tp < tpend && nec < nEC)
    {
      rp = *tp++;
      while (rp && nec < nEC)
	{
	  ec = rp->attribute->pure->EC;
	  if (ec != EC_UNDEF && inh_type[ec] == NULL)
	    {
#ifdef DEBUG1
	      printf("set EC %d type : \"%s\"\n",
		     ec, rp->attribute->pure->type);
#endif
	      inh_type[ec] = rp->attribute->pure->type;
	      nec++;
	    }
	  rp = rp->link;
	}
    }
}



/* output type decralations of local attributes.  */

static
void
output_local_type()
{
  register attribute_list **lp, **lpend, *list;
  register attribute *ap;
  
  if (nlocal_classes == 0)
      return;
  
#ifdef DEBUG1
  printf("outputing local types...");
#endif
  local_type = NEW2(nlocal_classes, attribute_list *);

  /* pack types  */
  pack_local_type();

  /* output type decralations  */
  fprintf(ftype, "typedef\n  struct rrlastype\n    {\n");

  lp = local_type;
  lpend = local_type + nlocal_classes;
  while (lp < lpend)
    {
      fprintf(ftype, "      union\n\t{\n");
      list = *lp++;
      if (list == NULL)
	internal("at output_local_type()");
      while (list)
	{
	  ap = list->attribute;
	  fprintf(ftype, "\t  %s\t%s_%s;\n",
		  ap->pure->type, ap->pure->name, ap->tag);
	  list = list->link;
	}
      fprintf(ftype, "\t} C%d;\n\n", ap->pure->EC);
    }

  fprintf(ftype, "    }\n  rrlastype;\n\n");

  FREE(local_type);

#ifdef DEBUG1
  printf("finish.\n");
#endif
}
    


/* pack types of local attributes into 'local_type[]'.  */

static
void
pack_local_type()
{
  register semantic_rule **tp, **tpend, *rp;
  register short class;
  register attribute_list *lp;
  
  tp = local_table;
  tpend = local_table + nitems;

  while (tp < tpend)
    {
      rp = *tp++;
      while (rp)
	{
	  class = rp->attribute->pure->EC;
	  if (class != EC_UNDEF)
	    {
	      lp = NEW(attribute_list);
	      lp->attribute = rp->attribute;
	      lp->link = local_type[class];
	      local_type[class] = lp;
	    }
	  rp = rp->link;
	}
    }
}



/* output semantic rules of synthesized attributes.  */

static
void
output_synthesize()
{
  register int ruleno;

  if (nsynthesizes == 0)
    return;

#ifdef DEBUG1
  printf("outputing semantic rules of synthesized attribute...");
#endif
  fprintf(freduction, "  switch (yyn)\n    {\n");

  ruleno = 1;
  for (; ruleno<= nrules; ++ruleno)
    {
      if (synt_table[ruleno]
	  ||
	  local_table[rrhsend[ruleno]]
	  ||
	  condition_table[ruleno])
	{
	  semantic_rule *rp;
	  semantic_condition *cp;
	  
	  fprintf(freduction, "      case %d:\n", ruleno);

	  /* output 'local_table[]'  */
	  rp = local_table[rrhsend[ruleno]];
	  for (; rp; rp=rp->link)
	    {
	      /* output "#line" pragma  */
	      if (!nolinesflag)
		fprintf(freduction, "#line %d \"%s\"\n", rp->lineno, infile);
  
	      putc('\t', freduction);
	      output_semrule(freduction, rp);
	      putc(';', freduction);
	      putc('\n', freduction);
	    }

	  /* output 'synt_table[]'  */
	  rp = synt_table[ruleno];
	  for (; rp; rp=rp->link)
	    {
	      /* output "#line" pragma  */
	      if (!nolinesflag)
		fprintf(freduction, "#line %d \"%s\"\n", rp->lineno, infile);
  
	      putc('\t', freduction);
	      output_semrule(freduction, rp);
	      putc(';', freduction);
	      putc('\n', freduction);
	    }

	  /* output 'condition_table[]'  */
	  cp = condition_table[ruleno];
	  for (; cp; cp=cp->link)
	    {
	      /* output "#line" pragma  */
	      if (!nolinesflag)
		fprintf(freduction, "#line %d \"%s\"\n", cp->lineno, infile);

	      output_semantic_condition(freduction, cp);
	    }
	  
	  fprintf(freduction, "\tbreak;\n\n");
	}
    }

  fprintf(freduction, "    }\n");
#ifdef DEBUG1
  printf("finish.\n");
#endif
}



/* output macros for parser  */

static
void
output_attribute_defines()
{
  /* define/undefine 'RIE_SOME_SYNTHESIZE' macro.  */
  if (nsynthesizes == 0)
    {
      fprintf(ftype, "#ifdef RIE_SOME_SYNTHESIZE\n");
      fprintf(ftype, "#undef RIE_SOME_SYNTHESIZE\n");
      fprintf(ftype, "#endif\n\n");
    }
  else
    {
      fprintf(ftype, "#ifndef RIE_SOME_SYNTHESIZE\n");
      fprintf(ftype, "#define RIE_SOME_SYNTHESIZE\n");
      fprintf(ftype, "#endif\n\n");
    }
  
  /* define/undefine 'RIE_SOME_INHERIT' macro.  */
  if (nEC == 0)
    {
      fprintf(ftype, "#ifdef RIE_SOME_INHERIT\n");
      fprintf(ftype, "#undef RIE_SOME_INHERIT\n");
      fprintf(ftype, "#endif\n\n");
    }
  else
    {
      fprintf(ftype, "#ifndef RIE_SOME_INHERIT\n");
      fprintf(ftype, "#define RIE_SOME_INHERIT\n");
      fprintf(ftype, "#endif\n\n");
    }
  
  /* define/undefine 'RIE_SOME_LOCAL' macro.  */
  if (nlocal_classes == 0)
    {
      fprintf(ftype, "#ifdef RIE_SOME_LOCAL\n");
      fprintf(ftype, "#undef RIE_SOME_LOCAL\n");
      fprintf(ftype, "#endif\n\n");
    }
  else
    {
      fprintf(ftype, "#ifndef RIE_SOME_LOCAL\n");
      fprintf(ftype, "#define RIE_SOME_LOCAL\n");
      fprintf(ftype, "#endif\n\n");
    }
}



/* output header of inherited attriubtes.  */

void
output_inherit_header()
{
    fprintf(fshift, "  switch (yystate)\n    {\n");
}



/* output trailer of inherited attributes.  */

void
output_inherit_trailer()
{
    fprintf(fshift, "    }\n\n");
}



/* output preceding code.  */

void
output_inherit_pre(stateno)
     int stateno;
{
  fprintf(fshift, "      case %d:\n", stateno);
}



/* output post code.  */

void
output_inherit_post()
{
  fprintf(fshift, "\tbreak;\n\n");
}



/* output semantic rules of inherited attributes.  */

void
output_inherit(rules)
     semantic_rule **rules;
{
  register semantic_rule **rp, **rpend;

  rpend = rules + nEC;

  rp = rules;
  while (rp < rpend)
    {
      if (*rp)
	{
	  /* output "#line" pragma  */
	  if (!nolinesflag)
	    fprintf(fshift, "#line %d \"%s\"\n", (*rp)->lineno, infile);
  
	  putc('\t', fshift);
	  output_semrule_plain(fshift, *rp);
	  putc(';', fshift);
	  putc('\n', fshift);
	}
      rp++;
    }
}



/* output semantic rules of local attributes  */

void
output_local(rlist)
     semantic_rule_list *rlist;
{
  while (rlist)
    {
      /* output "#line" pragma  */
      if (!nolinesflag)
	fprintf(fshift, "#line %d \"%s\"\n", rlist->rule->lineno, infile);
  
      putc('\t', fshift);
      output_semrule_plain(fshift, rlist->rule);
      putc(';', fshift);
      putc('\n', fshift);
      rlist = rlist->link;
    }
}



/* output #line pragma of action.  */

void
output_line_pragma()
{
  fprintf(faction, "#line %d \"%s\"\n", lineno, infile);
}



/* output semantic rules with parser format.  */

static
void
output_semrule(file, rule)
     FILE *file;
     semantic_rule *rule;
{
  /* output left hand side  */
  output_lhs_attribute(file, rule->attribute);

  /* output right hand side  */
  output_semexps(file, rule->expression, rule->attribute->offset);
}



/* output semantic rules with parser format.  */

static
void
output_semrule_plain(file, rule)
     FILE *file;
     semantic_rule *rule;
{
  /* output left hand side  */
  output_lhs_attribute(file, rule->attribute);

  /* output right hand side.  */
  output_semexps(file, rule->expression, 0);
}



/* output left hand side of semantic rule.  */

static
void
output_lhs_attribute(file, attr)
     FILE *file;
     attribute *attr;
{
  register pure *pp;

  pp = attr->pure;

  switch (pp->class)
    {
    case INHERIT:
      fprintf(file, "rriasp->EC%d =", pp->EC);
      break;

    case SYNTHESIZE:
      fprintf(file, "rrval.rr%s.%s =", attr->tag, pp->name);
      break;

    case LOCAL:
      fprintf(file, "rrlasp->C%d.%s_%s =", pp->EC, pp->name, attr->tag);
      break;

    default:
      internal("at output_lhs_attribute()");
      break;
    }
}



/* output expressions.  */

void
output_semexps(file, exp, base)
     FILE *file;
     semantic_expression *exp;
     int base;
{
  register attribute *ap;
  register pure *pp;
  
  while (exp)
    {
      switch (exp->kind)
	{
        case KCODE:
	  if(strcmp(exp->pointer.code, "new") == 0)     /* Yaman 92-02-10 */
            fprintf(file, "new ");                      /* Yaman 92-02-10 */
          else                                          /* Yaman 92-02-10 */
            fprintf(file, "%s", exp->pointer.code);
	  break;

	case KINHERIT:
	  ap = exp->pointer.attribute;
	  if (ap->offset - base == 0)
	    fprintf(file, " rriasp->EC%d", ap->pure->EC);
	  else
	    fprintf(file, " rriasp[%d].EC%d", ap->offset - base, ap->pure->EC);
	  break;

	case KSYNTHESIZE:
	  ap = exp->pointer.attribute;
	  if (ap->offset - base == 0)
	    fprintf(file, " rrsasp->rr%s.%s", ap->tag, ap->pure->name);
	  else
	    fprintf(file, " rrsasp[%d].rr%s.%s",
		    ap->offset - base, ap->tag, ap->pure->name);
	  break;

	case KLOCAL:
	  ap = exp->pointer.attribute;
	  pp = ap->pure;
	  if (ap->offset - base == 0)
	    fprintf(file, " rrlasp->C%d.%s_%s", pp->EC, pp->name, ap->tag);
	  else
	    fprintf(file, " rrlasp[%d].C%d.%s_%s",
		    ap->offset - base, pp->EC, pp->name, ap->tag);
	  break;

	case KSYNTHESIZE_LHS:
	  /* This is a special case.

	       We cannot distinguish the synthesized attribute of LHS from
	     the one of rightmost son.

	       Appearing this are not permitted while outputting inherited
	     or synthesized semantic rules.  But such a case will never
	     happen because check_L_attributed() reject the rules.

	       But in %condition, %message, %action or %debug, where we can
	     refer any attribute values even it is a synthesized attribute
	     of the LHS.  KSYNTHESIZE_LHS is used for distinguish them.

	       A synthesized attribute of Lhs. stored in 'rrval'

	   */
	  ap = exp->pointer.attribute;
	  fprintf(file, " rrval.rr%s.%s", ap->tag, ap->pure->name);
	  break;
	  
	default:
	  internal("at output_semexps()");
	  break;
	}

      exp = exp->next;
    }
}



/* output types of all attributes.  synthesized, inherited and local.  */

void
output_attribute_type()
{
  register int c;

  rewind(ftype);
  for (c=getc(ftype); c != EOF; c=getc(ftype))
    putc(c, ftable);
}


/* output the semantic condition.  */

void
output_semantic_condition(file, condp)
     FILE *file;
     semantic_condition *condp;
{
  switch (condp->kind)
    {
    case CCONDITION:
      fprintf(file, "\tif (!(");
      output_semexps(file, condp->expression, condp->base_offset);
      fprintf(file, "))\n");
      break;

    case CDEBUG:
      fprintf(file, "\tif (rrdebug)\n");
      /* through */

    case CACTION:
      fprintf(file, "\t");
      output_semexps(file, condp->expression, condp->base_offset);
      putc('\n', file);
      break;

    case CMESSAGE:
      fprintf(file, "\tfprintf(rrmessage, ");
      output_semexps(file, condp->expression, condp->base_offset);
      fprintf(file, ");\n");
      break;

    case CTERMINATOR:
    default:
      internal("at output_semantic_condition()");
    }
}
