/* print functions for Rie
   Copyright (C) 1991-1993 Masataka Sassa, Kazuhiro Kuroishi, Teruhisa Hirai.
   Copyright (C)      1994,1995                              Yoshiki Ohshima.
     
This file is part of Rie, A Compiler Compiler Based on ECLR-attributed Grammar.

Rie is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

Rie is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Bison; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

Bug reports, fixes or ideas are welcomed.
		Address: Yoshiki Ohshima
			 Tokyo Institute of Technology,
			 Department of Information Science,
			 Ookayama, Meguro-ku, Tokyo, 152 Japan.

		E-mail: rie-comments@is.titech.ac.jp			*/

#include <stdio.h>
#include "attribute.h"
#include "error.h"
#include "files.h"
#include "gram.h"
#include "new.h"

extern char **tags;
extern short * rrhsend;			/* outputsem.c  */
extern short *irule;			/* attribute.c  */
extern semantic_rule **inh_table;	/* attribute.c  */
extern semantic_rule **synt_table;	/* attribute.c  */
extern semantic_rule **local_table;	/* attribute.c  */
extern int nEC;				/* attribute.c  */
extern int nlocal_classes;		/* attribute.c  */

#define SEPARATE_MARK_CHAR	'$'


#if (defined(__STDC__) || defined(__GNUC__))

void print_EC_output(semantic_rule_list **);
void pirnt_local_output(semantic_rule_list *);
void print_terminator(void);
void print_EC_attributes(void);
void print_inherit(void);
void print_synthesize(int);
void print_cores(short *, short *);
void print_core_more(int);
static void print_semrule(FILE *, semantic_rule *);

#else	/* not ANSI C nor GNU C  */

void print_EC_output();
void print_local_output();
void print_terminator();
void print_EC_attributes();
void print_inherit();
void print_synthesize();
void print_cores();
void print_core_more();
static void print_semrule();

#endif	/* not ANSI C nor GNU C  */


/* write out semantic rules temporary to 'foutsub'.  */

void
print_EC_output(rlist)
     semantic_rule_list **rlist;
{
  register int i, firsttime;
  register semantic_rule_list *listp;
  
  for (i=0; i<nEC; i++)
    {
      firsttime = 0;
      
      listp = rlist[i];
      while (listp)
	{
	  if (firsttime == 0)
	    {
	      firsttime++;
	      fprintf(foutsub, "      EC %-2d", i);
	    }
	  else
	    fprintf(foutsub, "\t   ");

	  if (listp->flag == ECLR_VIOLATION)
	    putc('[', foutsub);
	  else
	    putc(' ', foutsub);
	  fprintf(foutsub, "{ ");

	  print_semrule(foutsub, listp->rule);

	  fprintf(foutsub, "; } (of rule %d)",
		  irule[listp->rule->attribute->offset]);
	  if (listp->flag == ECLR_VIOLATION)
	    putc(']', foutsub);
	  putc('\n', foutsub);
	  
	  listp = listp->link;
	}

    }	/* for  */

  putc('\n', foutsub);
}  


/* write semantic rules of local attribute to 'foutsub'.  */

void
print_local_output(rlist)
     semantic_rule_list *rlist;
{
  register int firsttime;
  
  firsttime = 0;
  
  while (rlist)
    {
      if (firsttime == 0)
	{
	  fprintf(foutsub, "      LOCAL { ");
	  firsttime++;
	}
      else
	fprintf(foutsub, "\t    { ");
      print_semrule(foutsub, rlist->rule);
      fprintf(foutsub, "; } (of rule %d)\n",
	      irule[rlist->rule->attribute->offset]);
      rlist = rlist->link;
    }
}


/* write out the termination character.  */

void
print_terminator()
{
  putc(SEPARATE_MARK_CHAR, foutsub);
}


/* print all attributes of each equivalence classes.  */

void
print_EC_attributes()
{
  register attribute_list **ECs, **ECs_tail, **lp, **lpend, *alist;
  register semantic_rule **tp, **tpend, *rp;
  register attribute *ap;
  register int ec, count;
  register char *tag, *name;
  
  fprintf(foutput, "\nEquivalence Classes :\n");

  ECs = NEW2(nEC, attribute_list *);
  ECs_tail = NEW2(nEC, attribute_list *);

  tp = inh_table;
  tpend = inh_table + nitems;

  while (tp < tpend)
    {
      rp = *tp++;
      while (rp)
	{
	  ap = rp->attribute;
	  ec = ap->pure->EC;
	  tag = ap->tag;
	  name = ap->pure->name;
	  
	  /* check if this attribute has already counted.  */
	  alist = ECs[ec];
	  while (alist)
	    {
	      ap = alist->attribute;
	      if (ap->tag == tag && ap->pure->name == name)
		break;
	      alist = alist->link;
	    }

	  if (alist == NULL)
	    {
	      /* new entry  */
	      alist = NEW(attribute_list);
	      alist->attribute = rp->attribute;
	      alist->link = NULL;
	      
	      if (ECs[ec] == NULL)
		ECs[ec] = alist;
	      else
		ECs_tail[ec]->link = alist;
	      ECs_tail[ec] = alist;
	    }
	  rp = rp->link;
	}	/* while(rp) */
    }	/* while(tp)  */

  /* write out  */
  ec = 0;
  lp = ECs;
  lpend = ECs + nEC;
  for (; lp < lpend; lp++)
    {
      if (*lp == NULL)
	continue;

      ap = (*lp)->attribute;
      fprintf(foutput, " class %d : %s.%s", ec++, ap->tag, ap->pure->name);

      count = 1;

      alist = (*lp)->link;
      while (alist)
	{
	  ap = alist->attribute;
	  fprintf(foutput, ", %s.%s", ap->tag, ap->pure->name);
	  if (++count > 3)
	    {
	      if (alist->link)
		fprintf(foutput, "\n\t   ");
	      count = 0;
	    }
	  alist = alist->link;
	}
      putc('\n', foutput);
    }

  FREE(ECs);
  FREE(ECs_tail);
}

  

/* put some contents of 'finherit' into 'foutput'.  */

void
print_inherit()
{
  register int c, column;

  column = 0;
  for (c=getc(foutsub); c != EOF; c=getc(foutsub))
    {
      if (column == 0 && c == SEPARATE_MARK_CHAR)
	break;
      putc(c, foutput);
      if (c == '\n')
	column = 0;
    }
}


/* print semantic rules at the reduction on each rule.  */

void
print_synthesize(ruleno)
     int ruleno;
{
  register int firsttime;
  register semantic_rule *rp;

  firsttime = 0;
  rp = local_table[rrhsend[ruleno]];

  while (rp) 
    {
      if (firsttime == 0)
	{
	  fprintf(foutput, "      LOCAL { ");
	  firsttime++;
	}
      else
	fprintf(foutput, "\t    { ");

      print_semrule(foutput, rp);

      fprintf(foutput, "; } (of rule %d)\n", ruleno);
      rp = rp->link;
    }

  firsttime = 0;
  rp = synt_table[ruleno];

  while (rp) 
    {
      if (firsttime == 0)
	{
	  fprintf(foutput, "      SYNT  { ");
	  firsttime++;
	}
      else
	fprintf(foutput, "\t    { ");

      print_semrule(foutput, rp);

      fprintf(foutput, "; } (of rule %d)\n", ruleno);
      rp = rp->link;
    }
}



void
print_cores(item, itemend)
     short *item;
     short *itemend;
{
  register int rule;
  register short *sp;
  register short *sp1;

  for (; item < itemend; item++)
    {
      rule = irule[*item];
      if (rule < 0)
	/* this 'rule' number is not used. see reduce_grammer_tables().  */
	continue;
      
      fprintf(foutsub, "    %s  ->  ", tags[rlhs[rule]]);

      sp  = ritem + rrhs[rule];
      sp1 = ritem + *item;
      while (sp < sp1)
	fprintf(foutsub, "%s ", tags[*sp++]);

      putc('.', foutsub);

      while (*sp > 0)
	fprintf(foutsub, " %s", tags[*sp++]);

      fprintf (foutsub, "   (rule %d)", rule);
      putc('\n', foutsub);
    }

  putc('\n', foutsub);
}


/* put some contents of 'foutsub' into 'foutput'.  */

void
print_core_more(state)
     int state;
{
  register int c, column;

  column = 0;
  for (c=getc(foutsub); c != EOF; c=getc(foutsub))
    {
      if (column == 0 && c == SEPARATE_MARK_CHAR)
	break;
      putc(c, foutput);
      if (c == '\n')
	column = 0;
    }
}


static
void
print_semrule(file, rule)
     FILE *file;
     semantic_rule *rule;
{
  register attribute *ap;
  register pure *pp;
  register semantic_expression *exp;
  
  ap = rule->attribute;
  pp = ap->pure;

  switch(pp->class)
    {
    case INHERIT:
    case SYNTHESIZE:
      if (ap->index == 1)
	fprintf(file, "%s.%s =", ap->tag, pp->name);
      else
	fprintf(file, "%s[%d].%s =", ap->tag, ap->index, pp->name);
      break;

    case LOCAL:
      fprintf(file, "%s = ", pp->name);
      break;

    default:
      internal("at print_semrule()");
    }

  exp = rule->expression;
  while (exp)
    {
      switch (exp->kind)
	{
	case KCODE:
	  fprintf(file, "%s", exp->pointer.code);
	  break;

	case KINHERIT:
	case KSYNTHESIZE:
	  ap = exp->pointer.attribute;
	  pp = ap->pure;
	  if (ap->index == 1)
	    fprintf(file, " %s.%s", ap->tag, pp->name);
	  else
	    fprintf(file, " %s[%d].%s", ap->tag, ap->index, pp->name);
	  break;

	case KLOCAL:
	  fprintf(file, " %s", exp->pointer.attribute->pure->name);
	  break;

	  default:
	  internal("at print_semexp()");
	}
      exp = exp->next;
    }
}
