/* Handle declared attributes
   Copyright (C) 1991-1993 Masataka Sassa, Kazuhiro Kuroishi, Teruhisa Hirai.
   Copyright (C)      1994,1995                              Yoshiki Ohshima.
     
This file is part of Rie, A Compiler Compiler Based on ECLR-attributed Grammar.

Rie is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

Rie is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Bison; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

Bug reports, fixes or ideas are welcomed.
		Address: Yoshiki Ohshima
			 Tokyo Institute of Technology,
			 Department of Information Science,
			 Ookayama, Meguro-ku, Tokyo, 152 Japan.

		E-mail: rie-comments@is.titech.ac.jp			*/

#include <stdio.h>
#include "new.h"
#include "pure.h"
#include "stringmgr.h"
#include "symtab.h"
#include "system.h"


#if (defined(__STDC__) || defined(__GNUC__))

pure *put_pure(bucket *, char *, char *, int);
pure *search_pure(bucket *, char *);
pure_list *copy_pure_lists(pure_list *);
void free_pure_lists(pure_list *);

#ifdef DEBUG
void print_pure(pure *);
void print_symbol_pures(bucket *);
#endif /* DEBUG */

#else	/* not ANSI C nor GNU C  */

pure *put_pure();
pure *search_pure();
pure_list *copy_pure_lists();
void free_pure_lists();

#ifdef DEBUG
void print_pure();
void print_symbol_pures();
#endif /* DEBUG */


#endif	/* nor ANSI C nor GNU C  */



pure *
put_pure(symbol, name, type, class)
     bucket *symbol;
     char *name;
     char *type;
     int class;
{
#ifdef DEBUG
  static int pure_no = 1;
#endif
  register pure_list *list;
  register pure *purep;
  
  list = symbol->pure_list;

  while (list)
    {
      if (strcmp(list->pure->name, name) == 0)
	return (list->pure);
      list = list->link;
    }

  /* new entry  */
  
#ifdef DEBUG
  printf("put attribute \"%s\" into symbol \"%s\" (NO %d)\n"
	 ,name, symbol->tag, pure_no);
  printf("  as class %s, type \"%s\"\n"
	 , (class == INHERIT) ? "INHERIT" : "SYNTHESIZE", type);
#endif	
  purep = NEW(pure);
  purep->name = copy_string(name);
  purep->type = copy_string(type);
  purep->class  = class;
  purep->EC = EC_UNDEF;
#ifdef DEBUG
  purep->no    = pure_no++;
#endif	

  list = NEW(pure_list);
  list->link = symbol->pure_list;
  list->pure = purep;
  
  symbol->pure_list = list;
  
  return (purep);
}




/* search and return the pure if symbol have it */

pure *
search_pure(symbol, name)
     bucket *symbol;
     char *name;
{
  register pure_list *list;
  
  list = symbol->pure_list;

  while (list)
    {
      if (strcmp(list->pure->name, name) == 0)
      	return (list->pure);
      list = list->link;
    }

  return NULL;
}



/* copy list of pure_list.  */

pure_list *
copy_pure_lists(list)
     pure_list *list;
{
  register pure_list *head, *tail, *lp;

  head = tail = NULL;
  
  while (list)
    {
      lp = NEW(pure_list);
      lp->pure = list->pure;
      lp->link = NULL;

      if (head == NULL)
	head = lp;
      else
	tail->link = lp;
      
      tail = lp;
      list = list->link;
    }

  return (head);
}


/* free list of pure_list  */

void
free_pure_lists(list)
     pure_list *list;
{
  register pure_list *next;

  while (list)
    {
      next = list->link;
      FREE(list);
      list = next;
    }
}



#ifdef DEBUG

/* libraries for debug */

void
print_pure(purep)
     pure *purep;
{
  printf("pure NO %d\n", purep->no);
  printf("     name   \"%s\"\n", purep->name ? purep->name : "");
  printf("     type   \"%s\"\n", purep->type ? purep->type : "");
  printf("     class  %s\n",
	 (purep->class == INHERIT) ? "INHERIT"
	 : (purep->class == SYNTHESIZE) ? "SYNTHESIZE"
	 : (purep->class == LOCAL) ? "LOCAL" : "unknown");
  printf("     EC     %d\n", purep->EC);
}


void
print_symbol_pures(symbol)
     bucket *symbol;
{
  register pure_list *listp;

  printf("\nAll attributes of symbol \"%s\"\n", symbol->tag);

  listp = symbol->pure_list;
  while (listp)
    {
      print_pure(listp->pure);
      listp = listp->link;
    }
}


#endif	/* DEBUG */
