/* A String Manager
   Copyright (C) 1991-1993 Masataka Sassa, Kazuhiro Kuroishi, Teruhisa Hirai.
   Copyright (C)      1994,1995                              Yoshiki Ohshima.
     
This file is part of Rie, A Compiler Compiler Based on ECLR-attributed Grammar.

Rie is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

Rie is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Bison; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

Bug reports, fixes or ideas are welcomed.
		Address: Yoshiki Ohshima
			 Tokyo Institute of Technology,
			 Department of Information Science,
			 Ookayama, Meguro-ku, Tokyo, 152 Japan.

		E-mail: rie-comments@is.titech.ac.jp			*/

#include <stdio.h>
#include "new.h"


/* 'INIT_TABLE_SIZE' must be a power of 2.  */
#define INIT_TABLE_SIZE		0x100


extern char *copys();	/* symtab.c  */


#if (defined(__STDC__) || defined(__GNUC__))

char *copy_string(char *);
void init_string_manager(void);
void final_string_manager(void);
static void expand_table(void);

#else	/* not ANSI C nor GNU C  */

char *copy_string();
void init_string_manager();
void final_string_manager();
static void expand_table();

#endif	/* not ANSI C nor GNU C  */


static char **string_table;
static char **table_end;
static unsigned table_size;
static unsigned occupy;
static unsigned table_size_mask;
static unsigned double_hashing_mask;



/* initialize string table  */

void
init_string_manager()
{
  table_size = (unsigned)INIT_TABLE_SIZE;
  table_size_mask = table_size - 1;
  double_hashing_mask = (table_size_mask >> 1);
  string_table = NEW2(table_size, char *);
  table_end = string_table + table_size;
  occupy = 0;
#ifdef DEBUG1
  printf("Initial string table.\n");
  printf("  table_size : 0x%x\n", table_size);
  printf("  table_size_mask : 0x%x\n", table_size_mask);
  printf("  double_hashing_mask : 0x%x\n", double_hashing_mask);
  printf("  occupy : 0x%x\n\n", occupy);
#endif
}


/* free string table  */

void
final_string_manager()
{
  FREE(string_table);
}



/* copy string  */

char *
copy_string(str)
     char *str;
{
  register char *cp, **hp;
  register int key;
  register int hash_addr, hash_inc;
  
  cp = str;
  key = 0;

  while (*cp)
    key += *cp++;

  /* hash_addr = (key mod table_size)  */
  hash_addr = (key & table_size_mask);

  if (string_table[hash_addr] == NULL)
    {
      /* empty  */
      cp = copys(str);
      string_table[hash_addr] = cp;

      if (++occupy >= table_size)
	expand_table();

#ifdef DEBUG1
      printf("copy_string() : new \"%s\" at %d\n", str, hash_addr);
#endif
      return cp;
    }

  /* conflict  */
  /* use Double Hashing algorithm  */

  /* hash_inc = (key mod table_size/2) * 2 + 1  */
  hash_inc  = ((key & double_hashing_mask) << 1) + 1;

  hp = &string_table[hash_addr];

  while (*hp && strcmp(*hp,str))
    {
      hp += hash_inc;
      if (hp >= table_end)
	hp -= table_size;
    }

  if (*hp)
    /* already registered  */
#ifdef DEBUG1
    {
      printf("copy_string() : \"%s\" at %d\n", str, hp - string_table);
#endif
    return *hp;
#ifdef DEBUG1
    }
#endif

  /* new entry  */
  *hp = copys(str);
  if (++occupy >= table_size)
    expand_table();

#ifdef DEBUG1
      printf("copy_string() : new \"%s\" at %d\n", str, hp - string_table);
#endif
  return *hp;
}



/* expand string table  */

static
void
expand_table()
{
  register char **old, **oldend, **tp;

#ifdef DEBUG1
  printf("\nExpanding string table...\n");
#endif

  double_hashing_mask = table_size_mask;

  /* table_size_mask = table_size_mask * 2 + 1  */
  table_size_mask |= table_size;

  old = string_table;
  oldend = string_table + table_size;

  /* table_size *= 2  */
  table_size <<= 1;

  string_table = NEW2(table_size, char *);
  table_end = string_table + table_size;
  occupy = 0;
  
  tp = old;
  while (tp < oldend)
    {
      if (*tp)
	copy_string(*tp);
      tp++;
    }

  FREE(old);

#ifdef DEBUG1
  printf("done.\n");
  printf("  table_size : 0x%x\n", table_size);
  printf("  table_size_mask : 0x%x\n", table_size_mask);
  printf("  double_hashing_mask : 0x%x\n", double_hashing_mask);
  printf("  occupy : 0x%x\n\n", occupy);
#endif
}

